"""Module for detection non-wear times from raw acceleration signals."""
import datetime
from typing import Tuple, Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_has_columns
from biopsykit.utils._types import arr_t
from biopsykit.utils.array_handling import sliding_window


class WearDetection:
    """Detect non-wear times from raw acceleration signals.

    Non-wear times are estimated over 15 minute intervals over the day.

    The non-wear detection was implemented according to [1] and modified according to [2].


    References
    ----------
    [1] Van Hees, V. T., Renström, F., Wright, A., Gradmark, A., Catt, M., Chen, K. Y., Löf, M., Bluck, L.,
    Pomeroy, J., Wareham, N. J., Ekelund, U., Brage, S., & Franks, P. W. (2011). Estimation of Daily Energy
    Expenditure in Pregnant and Non-Pregnant Women Using a Wrist-Worn Tri-Axial Accelerometer. PLoS ONE, 6(7), e22922.
    https://doi.org/10.1371/journal.pone.0022922

    [2] van Hees, V. T., Gorzelniak, L., Taherian, S., & Ekelund, U. (2013). Separating Movement and Gravity Components
    in an Acceleration Signal and Implications for the Assessment of Human Daily Physical Activity. PLOS ONE, 8(4), 10.
    https://doi.org/doi.org/10.1371/journal.pone.0061691

    """

    sampling_rate: float

    def __init__(self, sampling_rate: float):
        """Initialize a new ``WearDetection`` instance.

        Parameters
        ----------
        sampling_rate : float
            sampling rate of recorded data in Hz

        """
        self.sampling_rate = sampling_rate

    def predict(self, data: arr_t) -> pd.DataFrame:
        """Predict non-wear times from acceleration data.

        Parameters
        ----------
        data : array_like
            input acceleration data. Must be 3-d.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with wear (1) and non-wear (0) times per 15 minute interval

        """
        index = None
        index_resample = None
        if isinstance(data, (pd.DataFrame, pd.Series)):
            index = data.index

        data = data.filter(like="acc") if isinstance(data, pd.DataFrame) else pd.DataFrame(data)

        window = 60  # min
        overlap = 15  # min
        overlap_percent = 1.0 - (overlap / window)

        acc_sliding = {
            col: sliding_window(
                data[col].values,
                window_sec=window * 60,
                sampling_rate=self.sampling_rate,
                overlap_percent=overlap_percent,
            )
            for col in data
        }

        if index is not None:
            index_resample = self._resample_index(index, window, overlap_percent)

        acc_std = pd.DataFrame({axis: np.nanstd(acc_sliding[axis], ddof=1, axis=1) for axis in acc_sliding})

        acc_std[acc_std >= 0.013] = 1
        acc_std[acc_std < 0.013] = 0
        acc_std = np.nansum(acc_std, axis=1)

        acc_range = pd.DataFrame(
            {axis: np.nanmax(acc_sliding[axis], axis=1) - np.nanmin(acc_sliding[axis], axis=1) for axis in acc_sliding}
        )

        acc_range[acc_range >= 0.15] = 1
        acc_range[acc_range < 0.15] = 0
        acc_range = np.nansum(acc_range, axis=1)

        wear = np.ones(shape=acc_std.shape)
        wear[np.logical_or(acc_std < 1.0, acc_range < 1.0)] = 0.0

        wear = pd.DataFrame(wear, columns=["wear"])
        if index_resample is not None:
            wear = wear.join(index_resample)

        # apply rescoring three times
        wear = self._rescore_wear_detection(wear)
        wear = self._rescore_wear_detection(wear)
        wear = self._rescore_wear_detection(wear)

        return wear

    def _resample_index(self, index: pd.Index, window: int, overlap_percent: float):
        index_resample = sliding_window(
            np.arange(0, len(index)),
            window_sec=window * 60,
            sampling_rate=self.sampling_rate,
            overlap_percent=overlap_percent,
        )[:, :]
        start_end = index_resample[:, [0, -1]]
        if np.isnan(start_end[-1, -1]):
            last_idx = index_resample[-1, np.where(~np.isnan(index_resample[-1, :]))[0][-1]]
            start_end[-1, -1] = last_idx

        start_end = start_end.astype(int)

        if isinstance(index, pd.DatetimeIndex):
            index_resample = pd.DataFrame(index.to_numpy()[start_end], columns=["start", "end"])
        return index_resample

    @staticmethod
    def _rescore_wear_detection(data: pd.DataFrame) -> pd.DataFrame:
        # group classifications into wear and non-wear blocks
        data["block"] = data["wear"].diff().ne(0).cumsum()
        blocks = list(data.groupby("block"))

        # iterate through blocks
        for (_, prev), (idx_curr, curr), (_, post) in zip(blocks[0:-2], blocks[1:-1], blocks[2:]):
            if curr["wear"].unique():
                # get hour lengths of the previous, current, and next blocks
                dur_prev, dur_curr, dur_post = (len(dur) * 0.25 for dur in [prev, curr, post])

                if (dur_curr < 3 and dur_curr / (dur_prev + dur_post) < 0.8) or (
                    dur_curr < 6 and dur_curr / (dur_prev + dur_post) < 0.3
                ):
                    # a) if the current block is less than 3 hours and the ratio to previous and post blocks is
                    # less than 80% rescore the wear period as non-wear
                    # b) if the current block is less than 6 hours and the ratio to previous and post blocks is
                    # less than 30% rescore the wear period as non-wear
                    data.loc[data["block"] == idx_curr, "wear"] = 0

        data = data.drop(columns=["block"])
        return data

    @staticmethod
    def get_major_wear_block(data: pd.DataFrame) -> Tuple[Union[datetime.datetime, int], Union[datetime.datetime, int]]:
        """Return major wear block.

        The major wear block is the longest continuous wear block in the data.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            data with wear detection applied. The dataframe is expected to have a "wear" column.

        Returns
        -------
        start : :class:`~datetime.datetime` or int
            start of major wear block as datetime or int index
        end : :class:`~datetime.datetime` or int
            end of major wear block as datetime or int index

        See Also
        --------
        :meth:`~biopsykit.signals.imu.wear_detection.WearDetection.predict`
            apply wear detection on accelerometer data

        """
        data = data.copy()
        _assert_has_columns(data, [["wear"]])

        data["block"] = data["wear"].diff().ne(0).cumsum()
        wear_blocks = list(data.groupby("block").filter(lambda x: (x["wear"] == 1.0).all()).groupby("block"))
        max_block = wear_blocks[np.argmax([len(b) for i, b in wear_blocks])][1]
        max_block = (max_block["start"].iloc[0], max_block["end"].iloc[-1])
        return max_block

    @staticmethod
    def cut_to_wear_block(
        data: pd.DataFrame, wear_block: Tuple[Union[datetime.datetime, int], Union[datetime.datetime, int]]
    ) -> pd.DataFrame:
        """Cut data to wear block.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data that contains wear block
        wear_block : tuple
            tuple with start and end times of wear block. The type of ``wear_block`` depends on the index of ``data``.
            (datetime or int)

        Returns
        -------
        :class:`~pandas.DataFrame`
            data cut to wear block

        """
        if isinstance(data.index, pd.DatetimeIndex):
            return data.loc[wear_block[0] : wear_block[-1]]
        return data.iloc[wear_block[0] : wear_block[-1]]



"""Module for processing saliva data and computing established features (AUC, slope, maximum increase, ...)."""
from biopsykit.saliva import utils
from biopsykit.saliva.saliva import auc, initial_value, max_increase, max_value, mean_se, slope, standard_features

__all__ = ["auc", "initial_value", "max_increase", "max_value", "mean_se", "slope", "standard_features", "utils"]



"""Module with functions for model selection using "nested" cross-validation."""
import warnings
from typing import Any, Dict, Optional

import numpy as np
from biopsykit.classification.utils import split_train_test
from sklearn.metrics import confusion_matrix, get_scorer
from sklearn.model_selection import BaseCrossValidator, GridSearchCV, RandomizedSearchCV
from sklearn.pipeline import Pipeline
from tqdm.auto import tqdm

__all__ = ["nested_cv_param_search"]


def nested_cv_param_search(  # pylint:disable=invalid-name # pylint:disable=too-many-branches
    X: np.ndarray,  # pylint:disable=invalid-name  # noqa: N803
    y: np.ndarray,
    param_dict: Dict[str, Any],
    pipeline: Pipeline,
    outer_cv: BaseCrossValidator,
    inner_cv: BaseCrossValidator,
    groups: Optional[np.ndarray] = None,
    hyper_search_params: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    """Perform a cross-validated parameter search with hyperparameter optimization within a outer cross-validation.

    Parameters
    ----------
    X : array-like of shape (`n_samples`, `n_features`)
        Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.
    y : array-like of shape (`n_samples`, `n_output`) or (`n_samples`,)
        Target (i.e., class labels) relative to X for classification or regression.
    param_dict : dict or list of dicts
        Dictionary with parameters names (`str`) as keys and lists of parameter settings to try as values,
        or a list of such dictionaries, in which case the grids spanned by each dictionary in the list are explored.
        This enables searching over any sequence of parameter settings.
    pipeline : :class:`~sklearn.pipeline.Pipeline`
        Pipeline of sklearn transforms and estimators to perform hyperparameter search with.
    outer_cv : `CV splitter <https://scikit-learn.org/stable/glossary.html#term-CV-splitter>`_
        Cross-validation object determining the cross-validation splitting strategy of the outer cross-validation.
    inner_cv : `CV splitter`_
        Cross-validation object determining the cross-validation splitting strategy of the hyperparameter search.
    groups : array-like of shape (`n_samples`,)
        Group labels for the samples used while splitting the dataset into train/test set. Only used in conjunction
        with a "Group"``cv`` instance (e.g., :class:`~sklearn.model_selection.GroupKFold`).
        Default: ``None``
    hyper_search_params : dict, optional
        Dictionary specifying which hyperparameter search method to use (or ``None`` to use grid-search).

            * "grid" (:class:`~sklearn.model_selection.GridSearchCV`): To perform a grid-search pass a dict in the form
              of ``{"search_method": "grid"}``.
            * "random" (:class:`~sklearn.model_selection.RandomizedSearchCV`): To perform a randomized-search pass a
              dict in the form of ``{"search_method": "random", "n_iter": xx}``, where ``"n_iter"`` corresponds to the
              number of parameter settings that are sampled.

    kwargs : Additional arguments to be passed to the hyperparameter search class instance
             (e.g., :class:`~sklearn.model_selection.GridSearchCV` or
             :class:`~sklearn.model_selection.RandomizedSearchCV`).


    Returns
    -------
    dict
        Dictionary with hyperparameter search results. The result dictionary has the following entries:

        - "param_search": list with hyperparameter search class instances
          (e.g., :class:`~sklearn.model_selection.GridSearchCV`) used for hyperparameter search for each outer fold
          (determined by ``outer_cv``).
        - "test_score":  list with test scores of the best estimator on the respective test set for each outer fold.
        - "cv_results": list of ``cv_results_`` attributes of hyperparameter search class
          (e.g., :class:`~sklearn.model_selection.GridSearchCV`).
          Each entry of "cv_results" is a results dictionary of the respective fold with keys as column headers and
          values as columns, that can be imported into a pandas DataFrame.
        - "best_estimator" list of ``best_estimator_`` attributes of hyperparameter search class
          (e.g., :class:`~sklearn.model_selection.GridSearchCV`). Each entry of "best_estimator" is the estimator that
          was chosen by the hyperparameter in the respective fold, i.e. the estimator which gave the highest
          average score (or smallest loss if specified) on the test data.
        - "conf_matrix": list of confusion matrices from test scores for each outer fold

    See Also
    --------
    :class:`~sklearn.model_selection.GridSearchCV`
        sklearn grid-search
    :class:`~sklearn.model_selection.RandomizedSearchCV`
        sklearn randomized-search

    """
    if hyper_search_params is None:
        hyper_search_params = {"search_method": "grid"}

    scoring = kwargs.pop("scoring", None)
    kwargs, scoring_dict = _setup_scoring_dict(scoring, **kwargs)

    cols = [
        "param_search",
        "cv_results",
        "best_estimator",
        "conf_matrix",
        "predicted_labels",
        "true_labels",
        "train_indices",
        "test_indices",
    ]
    for scorer in scoring_dict:
        cols.append(f"test_{scorer}")
    results_dict = {key: [] for key in cols}

    # fix random states of cv objects for reproducibility
    if hasattr(outer_cv, "random_state"):
        outer_cv.random_state = kwargs.get("random_state", None)
    if hasattr(inner_cv, "random_state"):
        inner_cv.random_state = kwargs.get("random_state", None)

    for train, test in tqdm(list(outer_cv.split(X, y, groups)), desc="Outer CV"):
        cv_obj = _get_param_search_cv_object(
            pipeline, param_dict, inner_cv, scoring_dict, hyper_search_params, **kwargs
        )

        if groups is None:
            x_train, x_test, y_train, y_test = split_train_test(X, y, train, test)
            groups_train = None
        else:
            (  # pylint:disable=unbalanced-tuple-unpacking
                x_train,
                x_test,
                y_train,
                y_test,
                groups_train,
                _,
            ) = split_train_test(X, y, train, test, groups)

        cv_obj = _fit_cv_obj_one_fold(cv_obj, x_train, y_train, groups_train)

        results_dict["param_search"].append(cv_obj)
        for scorer in scoring_dict:
            results_dict[f"test_{scorer}"].append(get_scorer(scorer)._score_func(y_test, cv_obj.predict(x_test)))
        results_dict["train_indices"].append(train)
        results_dict["test_indices"].append(test)
        results_dict["predicted_labels"].append(cv_obj.predict(x_test))
        results_dict["true_labels"].append(y_test)
        results_dict["cv_results"].append(cv_obj.cv_results_)
        results_dict["best_estimator"].append(cv_obj.best_estimator_)
        try:
            results_dict["conf_matrix"].append(confusion_matrix(y_test, cv_obj.predict(x_test), normalize=None))
        except ValueError as e:
            if "Classification metrics can't handle a mix of multiclass and continuous targets" in e.args[0]:
                # TODO: add "error_handling" parameter to allow for different error handling strategies
                warnings.warn("Cannot compute confusion matrix for regression tasks.")

    return results_dict


def _setup_scoring_dict(scoring, **kwargs):
    scoring_dict = {}
    if scoring is not None:
        if isinstance(scoring, str):
            kwargs["refit"] = scoring
            scoring = [scoring]

        for score in scoring:
            scoring_dict.setdefault(score, score)
    return kwargs, scoring_dict


def _fit_cv_obj_one_fold(cv_obj, x_train, y_train, groups_train):
    try:
        if groups_train is None:
            cv_obj.fit(x_train, y_train)
        else:
            cv_obj.fit(x_train, y_train, groups=groups_train)
    except ValueError as e:
        if "Classification metrics can't handle a mix of multiclass and continuous targets" in e.args[0]:
            raise ValueError(
                "Error when attempting to fit estimator. "
                "It seems that you are trying to fit a regression model, "
                "but specified metrics for classification. "
                "Please check your code and provide other evaluation metrics if necessary!"
            ) from e
        if "An empty dict was passed." in e.args[0]:
            raise ValueError("No scoring metric was specified for the estimator!") from e
        raise ValueError from e
    return cv_obj


def _get_param_search_cv_object(
    pipeline: Pipeline,
    param_dict: Dict[str, Any],
    inner_cv: BaseCrossValidator,
    scoring_dict: Dict[str, str],
    hyper_search_config: Dict[str, Any],
    **kwargs,
):
    random_state = kwargs.pop("random_state", None)
    if hyper_search_config["search_method"] == "random":
        return RandomizedSearchCV(
            pipeline,
            param_distributions=param_dict,
            cv=inner_cv,
            scoring=scoring_dict,
            n_iter=hyper_search_config["n_iter"],
            random_state=random_state,
            **kwargs,
        )
    if hyper_search_config["search_method"] == "grid":
        return GridSearchCV(pipeline, param_grid=param_dict, cv=inner_cv, scoring=scoring_dict, **kwargs)
    raise ValueError("Unknown search method {}".format(hyper_search_config["search_method"]))



"""Module containing various utility functions (time, dataframe, array, file handling, ...)."""
from biopsykit.utils import (
    array_handling,
    data_processing,
    dataframe_handling,
    datatype_helper,
    exceptions,
    file_handling,
    functions,
    legacy_helper,
    time,
)

__all__ = [
    "time",
    "functions",
    "file_handling",
    "datatype_helper",
    "array_handling",
    "dataframe_handling",
    "exceptions",
    "data_processing",
    "legacy_helper",
]



"""A set of custom exceptions."""

__all__ = [
    "ValidationError",
    "FileExtensionError",
    "DataFrameTransformationError",
    "ValueRangeError",
    "FeatureComputationError",
    "EcgProcessingError",
]


class ValidationError(Exception):
    """An error indicating that data-object does not comply with the guidelines."""


class FileExtensionError(Exception):
    """An error indicating that the file name has the wrong file extension."""


class DataFrameTransformationError(Exception):
    """An error indicating that dataframe transformation failed."""


class ValueRangeError(Exception):
    """An error indicating that scores are not in the expected range."""


class FeatureComputationError(Exception):
    """An error indicating that feature computation failed."""


class EcgProcessingError(Exception):
    """An error indicating that ECG processing failed."""



"""Module containing implementations to compute various metadata parameter.

Each function at least expects a dataframe containing the required columns in a specified order
(see function documentations for specifics) to be passed to the ``data`` argument.

If ``data`` is a dataframe that contains more than the required two columns, e.g., if the complete dataframe
is passed, the required columns can be sliced by specifying them in the ``columns`` parameter.

"""
from typing import Optional, Sequence, Union

import pandas as pd
from biopsykit.utils._datatype_validation_helper import (
    _assert_has_columns,
    _assert_has_index_levels,
    _assert_len_list,
    _assert_value_range,
)

__all__ = ["bmi", "whr", "gender_counts"]

from biopsykit.utils.exceptions import ValidationError


def gender_counts(
    data: pd.DataFrame, gender_col: Optional[str] = None, split_condition: Optional[bool] = False
) -> pd.DataFrame:
    """Get statistics about gender distribution from a dataset.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with subjects
    gender_col : str, optional
        column name containing gender information or ``None`` to use default name ("gender").
    split_condition : bool, optional
        ``True`` to split gender distribution by condition (assumes that an "condition" index level is present in
        ``data``), ``False`` otherwise.
        Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with absolute and relative gender distribution

    """
    if gender_col is None:
        gender_col = "gender"
    if isinstance(gender_col, str):
        gender_col = [gender_col]

    try:
        _assert_len_list(gender_col, 1)
    except ValidationError as e:
        raise ValidationError(
            f"'gender_col' is excepted to be only one column! Got {len(gender_col)} columns instead."
        ) from e
    _assert_has_columns(data, [gender_col])
    data = data.loc[:, gender_col]
    if split_condition:
        _assert_has_index_levels(data, "condition", match_atleast=True)
        return data.groupby("condition").apply(_gender_counts)
    return _gender_counts(data)


def _gender_counts(data: pd.DataFrame):
    return pd.concat(
        [data.value_counts(sort=False), data.value_counts(normalize=True, sort=False) * 100],
        axis=1,
        keys=["count", "percent"],
    )


def bmi(data: pd.DataFrame, columns: Optional[Union[Sequence[str], pd.Index]] = None) -> pd.DataFrame:
    """Compute the **Body Mass Index**.

    This function assumes the required data in the following format:

    * 1st column: weight in kilogram
    * 2nd column: height in centimeter

    If ``data`` is a dataframe that contains more than the required two columns, e.g., if the complete questionnaire
    dataframe is passed, the required columns can be sliced by specifying them in the ``columns`` parameter.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe containing weight and height information
    columns: sequence of str, optional
        list of column names needed to compute body mass index. Only needed if ``data`` is a dataframe with more than
        the required columns for computing body mass index. Not needed if ``data`` only contains the required columns.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with body mass index

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValueRangeError`
        if input values or output values are not in the expected range, e.g., because values are provided in the
        wrong unit or columns are in the wrong order

    """
    score_name = "BMI"

    if columns is not None:
        # if columns parameter is supplied: slice columns from dataframe
        data = data.loc[:, columns]

    # weight
    _assert_value_range(data.iloc[:, 0], [10, 300])
    # height
    _assert_value_range(data.iloc[:, 1], [50, 250])

    data = pd.DataFrame(data.iloc[:, 0] / (data.iloc[:, 1] / 100.0) ** 2, columns=[score_name])
    # check if BMI is in a reasonable range
    _assert_value_range(data, [10, 50])
    return data.round(2)


def whr(data: pd.DataFrame, columns: Optional[Union[Sequence[str], pd.Index]] = None) -> pd.DataFrame:
    """Compute the **Waist to Hip Ratio**.

    This function assumes the required data in the following format:

    * 1st column: waist circumference
    * 2nd column: hip circumference

    If ``data`` is a dataframe that contains more than the required two columns, e.g., if the complete questionnaire
    dataframe is passed, the required columns can be sliced by specifying them in the ``columns`` parameter.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe containing waist and hip circumference
    columns: sequence of str, optional
        list of column names needed to compute body mass index. Only needed if ``data`` is a dataframe with more than
        the required columns for computing body mass index. Not needed if ``data`` only contains the required columns.
        Default: ``None``

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with waist to hip ratio

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValueRangeError`
        if input values or output values are not in the expected range, e.g., because values are provided in the
        wrong unit or column are in the wrong order

    """
    score_name = "WHR"

    if columns is not None:
        # if columns parameter is supplied: slice columns from dataframe
        data = data.loc[:, columns]

    data = pd.DataFrame(data.iloc[:, 0] / data.iloc[:, 1], columns=[score_name])
    # check if WHR is in a reasonable range
    _assert_value_range(data, [0.5, 1.5])
    return data.round(3)



"""Module for importing data recorded by NilsPod sensors."""
import datetime
import re
import warnings
from pathlib import Path
from typing import Dict, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_is_dir, _assert_is_dtype
from biopsykit.utils._types import path_t
from biopsykit.utils.time import tz
from nilspodlib import Dataset, SyncedSession

COUNTER_INCONSISTENCY_HANDLING = Literal["raise", "warn", "ignore"]
"""Available behavior types when dealing with NilsPod counter inconsistencies."""

__all__ = [
    "load_dataset_nilspod",
    "load_synced_session_nilspod",
    "load_csv_nilspod",
    "load_folder_nilspod",
    "check_nilspod_dataset_corrupted",
    "get_nilspod_dataset_corrupted_info",
]


def load_dataset_nilspod(
    file_path: Optional[path_t] = None,
    dataset: Optional[Dataset] = None,
    datastreams: Optional[Union[str, Sequence[str]]] = None,
    handle_counter_inconsistency: Optional[COUNTER_INCONSISTENCY_HANDLING] = "raise",
    **kwargs,
) -> Tuple[pd.DataFrame, float]:
    """Load NilsPod recording and convert into dataframe.

    To load a dataset either a :class:`~nilspodlib.dataset.Dataset` object (via ``dataset`` parameter)
    or the path to the binary file (via ``file_path`` variable) can be passed.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str, optional
        path to binary file
    dataset : :class:`~nilspodlib.dataset.Dataset`, optional
        Dataset object
    datastreams : str or list of str, optional
        list of datastreams if only specific datastreams of the dataset object should be imported or
        ``None`` to load all datastreams. Datastreams that are not part of the current dataset will be silently ignored.
        Default: ``None``
    handle_counter_inconsistency : {"raise", "warn", "ignore"}, optional
        how to handle if counter of dataset is not monotonously increasing, which might be an indicator for a
        corrupted dataset:

        * "raise" (default): raise an error
        * "warn": issue a warning but still return a dataframe
        * "ignore": ignore the counter check result

    **kwargs
        optional arguments passed to :func:`~nilspodlib.Dataset.from_bin_file`, such as:

        legacy_support : {"error", "warn", "resolve"}, optional
            Flag indicating how to deal with older NilsPod firmware versions:

            * "error": raise an error if an unsupported version is detected
            * "warn": issue a warning and parse the file without modification
            * "resolve" (default): perform a legacy conversion to load old files. If no suitable conversion is found,
              an error is raised. See the :any:`nilspodlib.legacy` package and the README of ``nilspodlib``
              to learn more about available conversions.

        timezone : str or :class:`datetime.tzinfo`, optional
            timezone of the acquired data, either as string of as tzinfo object.
            Default: "Europe/Berlin"

    Returns
    -------
    tuple
        df : :class:`~pandas.DataFrame`
            dataframe of imported dataset
        fs : float
            sampling rate

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if `file_path` is specified and file is not a binary (.bin) file
    ValueError
        if neither `file_path` nor `dataset` are supplied as parameter, if ``handle_counter_inconsistency`` is
        ``raise`` and :class:`~nilspodlib.dataset.Dataset` counter is inconsistent (not monotonously increasing),
        if ``legacy_support`` is ``raise`` and so suitable conversion can be found for this file version.

    See Also
    --------
    :class:`~nilspodlib.dataset.Dataset`
        NilsPod Dataset

    Examples
    --------
    >>> from biopsykit.io.nilspod import load_dataset_nilspod
    >>> # Option 1: Import data by passing file name
    >>> file_path = "./<filename-of-nilspod-data>.bin"
    >>> # load dataset with all datastreams
    >>> df, fs = load_dataset_nilspod(file_path=file_path)
    >>> # load only ECG data of dataset
    >>> df, fs = load_dataset_nilspod(file_path=file_path, datastreams=['ecg'])
    >>>
    >>> # Option 2: Import data by passing Dataset object imported from NilsPodLib
    >>> # (in this example, only acceleration data)
    >>> from nilspodlib import Dataset
    >>> dataset = Dataset.from_bin_file("<filename>.bin")
    >>> df, fs = load_dataset_nilspod(dataset=dataset, datastreams='acc')

    """
    if file_path is not None:
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".bin")
        kwargs.setdefault("tz", kwargs.pop("timezone", tz))
        kwargs.setdefault("legacy_support", "resolve")
        dataset = Dataset.from_bin_file(file_path, **kwargs)

    if file_path is None and dataset is None:
        raise ValueError("Either 'file_path' or 'dataset' must be supplied as parameter!")

    _handle_counter_inconsistencies_dataset(dataset, handle_counter_inconsistency)

    if isinstance(datastreams, str):
        datastreams = [datastreams]

    # convert dataset to dataframe and localize timestamp
    df = dataset.data_as_df(datastreams, index="local_datetime")
    df.index.name = "time"
    return df, dataset.info.sampling_rate_hz


def load_synced_session_nilspod(
    folder_path: path_t,
    datastreams: Optional[Union[str, Sequence[str]]] = None,
    handle_counter_inconsistency: Optional[COUNTER_INCONSISTENCY_HANDLING] = "raise",
    **kwargs,
) -> Tuple[pd.DataFrame, float]:
    """Load a synchronized session of NilsPod recordings and convert into dataframes.

    Parameters
    ----------
    folder_path : :class:`~pathlib.Path` or str, optional
        folder path to session files
    datastreams : list of str, optional
        list of datastreams if only specific datastreams of the datasets in the session should be imported or
        ``None`` to load all datastreams. Datastreams that are not part of
        the current datasets will be silently ignored.
        Default: ``None``
    handle_counter_inconsistency : {"raise", "warn", "ignore"}, optional
        how to handle if counter of dataset is not monotonously increasing, which might be an indicator for a
        corrupted dataset:

        * "raise" (default): raise an error
        * "warn": issue a warning but still return a dataframe
        * "ignore": ignore the counter check result

    **kwargs
        optional arguments passed to :func:`~nilspodlib.Session.from_folder_path`, such as:

        legacy_support : {"error", "warn", "resolve"}, optional
            Flag indicating how to deal with older NilsPod firmware versions:

            * "error": raise an error if an unsupported version is detected
            * "warn": issue a warning and parse the file without modification
            * "resolve" (default): perform a legacy conversion to load old files. If no suitable conversion is found,
              an error is raised. See the :any:`nilspodlib.legacy` package and the README of ``nilspodlib``
              to learn more about available conversions.

        timezone : str or :class:`datetime.tzinfo`, optional
            timezone of the acquired data, either as string of as tzinfo object.
            Default: "Europe/Berlin"

    Returns
    -------
    tuple
        df : :class:`~pandas.DataFrame`
            concatenated dataframe of imported session
        fs : float
            sampling rate

    Raises
    ------
    ValueError
        if ``handle_counter_inconsistency`` is ``raise`` and counter of any dataset is inconsistent
        (not monotonously increasing).
        If ``legacy_support`` is ``raise`` and so suitable conversion can be found for the files in the session.
        If sampling rate is not the same for all datasets in the session
            ValueError
        If ``folder_path`` does not contain any NilsPod files
        If the sampling rates of the files in the folder are not the same

    See Also
    --------
    :class:`~nilspodlib.dataset.Dataset`
        NilsPod Dataset
    :func:`~biopsykit.io.nilspod.load_dataset_nilspod`
        load a single NilsPod dataset

    """
    # ensure pathlib
    folder_path = Path(folder_path)

    nilspod_files = sorted(folder_path.glob("*.bin"))
    if len(nilspod_files) == 0:
        raise ValueError("No NilsPod files found in directory!")

    kwargs.setdefault("tz", kwargs.pop("timezone", tz))
    session = SyncedSession.from_folder_path(folder_path, **kwargs)
    session = session.align_to_syncregion()

    _handle_counter_inconsistencies_session(session, handle_counter_inconsistency)
    if isinstance(datastreams, str):
        datastreams = [datastreams]

    # convert dataset to dataframe and localize timestamp
    df = session.data_as_df(datastreams, index="local_datetime", concat_df=True)
    df.index.name = "time"
    if len(set(session.info.sampling_rate_hz)) > 1:
        raise ValueError(
            f"Datasets in the sessions have different sampling rates! Got: {session.info.sampling_rate_hz}."
        )
    fs = session.info.sampling_rate_hz[0]
    return df, fs


def _handle_counter_inconsistencies_dataset(
    dataset: Dataset, handle_counter_inconsistency: COUNTER_INCONSISTENCY_HANDLING
):
    idxs_corrupted = np.where(np.diff(dataset.counter) < 1)[0]
    # edge case: check if only last sample is corrupted. if yes, cut last sample
    if len(idxs_corrupted) == 1 and (idxs_corrupted == len(dataset.counter) - 2):
        dataset = dataset.cut(start=0, stop=idxs_corrupted[0])
    elif len(idxs_corrupted) > 1:
        if handle_counter_inconsistency == "raise":
            raise ValueError("Error loading dataset. Counter not monotonously increasing!")
        if handle_counter_inconsistency == "warn":
            warnings.warn(
                "Counter not monotonously increasing. This might indicate that the dataset is corrupted or "
                "that the dataset was recorded as part of a synchronized session and might need to be loaded "
                "using `biopsykit.io.nilspod.load_synced_session_nilspod()`. "
                "Check the counter of the DataFrame manually!"
            )


def _handle_counter_inconsistencies_session(
    session: SyncedSession, handle_counter_inconsistency: COUNTER_INCONSISTENCY_HANDLING
):
    idxs_corrupted = np.where(np.diff(session.counter) < 1)[0]
    # edge case: check if only last sample is corrupted. if yes, cut last sample
    if len(idxs_corrupted) == 1 and (idxs_corrupted == len(session.counter) - 2):
        session = session.cut(start=0, stop=idxs_corrupted[0])
    elif len(idxs_corrupted) > 1:
        if handle_counter_inconsistency == "raise":
            raise ValueError("Error loading session. Counter not monotonously increasing!")
        if handle_counter_inconsistency == "warn":
            warnings.warn(
                "Counter not monotonously increasing. This might indicate that the session is corrupted. "
                "Check the counter of the DataFrame manually!"
            )


def load_csv_nilspod(
    file_path: path_t = None,
    datastreams: Optional[Sequence[str]] = None,
    timezone: Optional[Union[datetime.tzinfo, str]] = tz,
    filename_regex: Optional[str] = None,
    time_regex: Optional[str] = None,
) -> Tuple[pd.DataFrame, float]:
    r"""Convert a csv file recorded by NilsPod into a dataframe.

    By default, this function expects the file name to have the following pattern:
    "NilsPodX-<sensor-id>_YYYYMMDD_hhmmss.csv". The time information in the file name is used
    to infer the start time of the recording and add absolute time information to return
    a dataframe with a :class:`~pandas.DatetimeIndex`.

    If no start time can be extracted the index of the resulting
    dataframe is a :class:`~pandas.TimedeltaIndex`, not a :class:`~pandas.DatetimeIndex`.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str, optional
        path to binary file
    datastreams : list of str, optional
        list of datastreams if only specific datastreams of the file should be imported
        or ``None`` to load all datastreams. Datastreams that are not part of the current dataset will
        be silently ignored.
        Default: ``None``
    timezone : str or :class:`datetime.tzinfo`, optional
        timezone of the acquired data, either as string or as tzinfo object.
        Default: 'Europe/Berlin'
    filename_regex : str, optional
        regex string to extract time substring from file name or ``None`` to use default file name pattern.
        Default: ``None``
    time_regex : str, optional
        regex string specifying format of time substring or ``None`` to use default time format.
        Default: ``None``

    Returns
    -------
    df : :class:`~pandas.DataFrame`
        dataframe of imported dataset
    fs : float
        sampling rate

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no csv file

    See Also
    --------
    :class:`~nilspodlib.dataset.Dataset`
        NilsPod Dataset
    `load_dataset_nilspod`
        load a single NilsPod dataset from binary file

    """
    _assert_file_extension(file_path, ".csv")

    df = pd.read_csv(file_path, header=1, index_col="timestamp")
    header = pd.read_csv(file_path, header=None, nrows=1)

    # sampling rate is in second column of header
    sampling_rate = float(header.iloc[0, 1])

    if filename_regex is None:
        filename_regex = r"NilsPodX-[^\s]{4}_(.*?).csv"
    if time_regex is None:
        time_regex = "%Y%m%d_%H%M%S"

    # convert index to nanoseconds
    df.index = ((df.index / sampling_rate) * 1e9).astype(int)
    # infer start time from filename
    start_time = re.findall(filename_regex, str(file_path.name))
    df = _convert_index(df, start_time, time_regex)

    if isinstance(datastreams, str):
        datastreams = [datastreams]
    if datastreams is not None:
        # filter only desired datastreams
        df = pd.concat([df.filter(like=ds) for ds in datastreams], axis=1)

    if isinstance(df.index, pd.DatetimeIndex):
        # localize timezone (is already in correct timezone since start time is inferred from file name)
        df = df.tz_localize(tz=timezone)
    return df, sampling_rate


def _convert_index(df: pd.DataFrame, start_time: Sequence[str], time_regex: str):
    if len(start_time) > 0:
        # convert index to datetime index with absolute time information
        start_time = start_time[0]
        start_time = pd.to_datetime(start_time, format=time_regex).to_datetime64().astype(int)
        # add start time as offset and convert into datetime index
        df.index = pd.to_datetime(df.index + start_time)
    else:
        # no start time information available, so convert into timedelta index
        df.index = pd.to_timedelta(df.index)
    df.index.name = "time"

    return df


def load_folder_nilspod(
    folder_path: path_t, phase_names: Optional[Sequence[str]] = None, **kwargs
) -> Tuple[Dict[str, pd.DataFrame], float]:
    """Load all NilsPod datasets from one folder, convert them into dataframes, and combine them into a dictionary.

    This function can for example be used when single NilsPod sessions (datasets) were recorded
    for different study phases.

    Parameters
    ----------
    folder_path : :class:`~pathlib.Path` or str, optional
        folder path to files
    phase_names: list, optional
        list of phase names corresponding to the files in the folder. Must match the number of recordings.
        If ``None`` phase names will be named ``Part{1-x}``. Default: ``None``
    **kwargs
        additional arguments that are passed to :func:`load_dataset_nilspod`

    Returns
    -------
    dataset_dict : dict
        dictionary with phase names as keys and pandas dataframes with sensor recordings as values
    fs : float
        sampling rate of sensor recordings

    Raises
    ------
    ValueError
        if ``folder_path`` does not contain any NilsPod files, the number of phases does not match the number of
        datasets in the folder, or if the sampling rates of the files in the folder are not the same

    See Also
    --------
    :func:`load_dataset_nilspod`
        load single NilsPod dataset


    Examples
    --------
    >>> from biopsykit.io.nilspod import load_folder_nilspod
    >>> folder_path = "./nilspod"
    >>> # load all datasets from the selected folder with all datastreams
    >>> dataset_dict, fs = load_folder_nilspod(folder_path)
    >>> # load only ECG data of all datasets from the selected folder
    >>> dataset_dict, fs = load_folder_nilspod(folder_path, datastreams=['ecg'])
    >>> # load all datasets from the selected folder with correspondng phase names
    >>> dataset_dict, fs = load_folder_nilspod(folder_path, phase_names=['VP01','VP02','VP03'])

    """
    # ensure pathlib
    folder_path = Path(folder_path)
    _assert_is_dir(folder_path)

    # look for all NilsPod binary files in the folder
    dataset_list = sorted(folder_path.glob("*.bin"))
    if len(dataset_list) == 0:
        raise ValueError(f"No NilsPod files found in folder {folder_path}!")
    if phase_names is None:
        phase_names = [f"Part{i}" for i in range(len(dataset_list))]

    if len(phase_names) != len(dataset_list):
        raise ValueError(
            f"Number of phases does not match number of datasets in the folder! "
            f"Expected {len(dataset_list)}, got {len(phase_names)}."
        )

    dataset_list = [load_dataset_nilspod(file_path=dataset_path, **kwargs) for dataset_path in dataset_list]

    # check if sampling rate is equal for all datasets in folder
    fs_list = [fs for df, fs in dataset_list]

    if len(set(fs_list)) > 1:
        raise ValueError(f"Datasets in the sessions have different sampling rates! Got: {fs_list}.")
    fs = fs_list[0]

    dataset_dict = {phase: df for phase, (df, fs) in zip(phase_names, dataset_list)}
    return dataset_dict, fs


def check_nilspod_dataset_corrupted(dataset: Dataset) -> bool:
    """Check if a NilsPod dataset is potentially corrupted.

    A dataset is potentially corrupted if the counter is not monotonously increasing.

    Parameters
    ----------
    dataset : :class:`~nilspodlib.dataset.Dataset`
        dataset to check

    Returns
    -------
    bool
        flag indicating whether a NilsPod dataset is potentially corrupted or not

    """
    return np.where(np.diff(dataset.counter) != 1.0)[0].size != 0


def get_nilspod_dataset_corrupted_info(dataset: Dataset, file_path: path_t) -> Dict:
    """Get information about the corruption state of a NilsPod dataset.

    Corruption information include the information:

        * "name": recording date and time
        * "percent_corrupt": Amount of corrupted data in percent
        * "condition": Condition of the dataset. Can be one of:

          * "fine": if dataset is not corrupted
          * "lost": if more than 90% of all samples are corrupted
          * "parts": if between 50% and 90% of all samples are corrupted
          * "start_only": if less than 50% of all samples are corrupted and corrupted samples
            are only in the first third of the dataset
          * "end_only": if less than 50% of all samples are corrupted and corrupted samples
            are only in the last third of the dataset

    Parameters
    ----------
    dataset : :class:`~nilspodlib.dataset.Dataset`, optional
        Dataset object
    file_path : :class:`~pathlib.Path` or str, optional
        path to binary file

    Returns
    -------
    dict
        dictionary with corruption information

    """
    _assert_is_dtype(dataset, Dataset)
    nilspod_file_pattern = r"NilsPodX-\w{4}_(.*?).bin"
    # ensure pathlib
    file_path = Path(file_path)

    keys = ["name", "percent_corrupt", "condition"]
    dict_res = dict.fromkeys(keys)
    re_groups = re.search(nilspod_file_pattern, file_path.name)
    name = re_groups.group(1) if re_groups is not None else file_path.name
    dict_res["name"] = name
    if not check_nilspod_dataset_corrupted(dataset):
        dict_res["condition"] = "fine"
        dict_res["percent_corrupt"] = 0.0
        return dict_res

    idx_diff = np.diff(dataset.counter)
    idx_corrupt = np.where(idx_diff != 1.0)[0]
    percent_corrupt = round((len(idx_corrupt) / len(idx_diff)) * 100.0, 1)
    condition = _get_nilspod_dataset_corrupted_info_get_condition(percent_corrupt, idx_corrupt)

    dict_res["percent_corrupt"] = percent_corrupt
    dict_res["condition"] = condition
    return dict_res


def _get_nilspod_dataset_corrupted_info_get_condition(percent_corrupt: float, idx_corrupt: Sequence[int]) -> str:
    condition = "parts"
    if percent_corrupt > 90.0:
        condition = "lost"
    elif percent_corrupt < 50.0:
        if (idx_corrupt[0] / len(idx_corrupt)) < 0.30:
            condition = "start_only"
        elif (idx_corrupt[0] / len(idx_corrupt)) > 0.70:
            condition = "end_only"
    return condition



"""Module implementing a base class to represent psychological protocols."""
import json
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple, Type, Union

import biopsykit.protocols.plotting as plot
import matplotlib.pyplot as plt
import pandas as pd
from biopsykit.io import write_pandas_dict_excel
from biopsykit.protocols._utils import _check_sample_times_match, _get_sample_times
from biopsykit.signals.ecg import EcgProcessor
from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_is_dtype
from biopsykit.utils._types import T, path_t
from biopsykit.utils.data_processing import (
    add_subject_conditions,
    cut_phases_to_shortest,
    mean_per_subject_dict,
    mean_se_per_phase,
    merge_study_data_dict,
    normalize_to_phase,
    rearrange_subject_data_dict,
    resample_dict_sec,
    select_dict_phases,
    split_dict_into_subphases,
    split_subject_conditions,
)
from biopsykit.utils.datatype_helper import (
    HeartRateSubjectDataDict,
    SalivaFeatureDataFrame,
    SalivaRawDataFrame,
    SubjectDataDict,
    is_hr_subject_data_dict,
    is_saliva_mean_se_dataframe,
    is_saliva_raw_dataframe,
    is_subject_data_dict,
)
from biopsykit.utils.exceptions import ValidationError
from tqdm.auto import tqdm


class BaseProtocol:  # pylint:disable=too-many-public-methods
    """Base class representing a psychological protocol and data collected within a study."""

    def __init__(
        self,
        name: str,
        structure: Optional[Dict[str, Any]] = None,
        test_times: Optional[Sequence[int]] = None,
        **kwargs,
    ):
        """Class representing a base class for psychological protocols and data collected within a study.

        The general structure of the protocol can be specified by passing a ``structure`` dict to the constructor of
        ``BaseProtocol``.

        Up to three nested structure levels are supported:

        * 1st level: ``study part``: Different parts of the study, such as: "Preface", "Test",
          and "Questionnaires"
        * 2nd level: ``phase``: Different phases of the psychological protocol that belong to the same *study
          part*, such as: "Preparation", "Stress", "Recovery"
        * 3rd level: ``subphase``: Different subphases that belong to the same *phase*, such as:
          "Baseline", "Arithmetic Task", "Feedback"

        .. note::
            Duration of phases and/or subphases are expected in **seconds**.


        Parameters
        ----------
        name : str
            name of protocol
        structure : dict, optional
            nested dictionary specifying the structure of the protocol.
            Up to three nested structure levels are supported:

            * 1st level: ``study_part``: Different parts of the study, such as: "Preface", "Test",
              and "Questionnaires"
            * 2nd level: ``phase``: Different phases of the psychological protocol that belong to the same *study
              part*, such as: "Preparation", "Stress", "Recovery"
            * 3rd level: ``subphase``: Different subphases that belong to the same *phase*, such as:
              "Baseline", "Arithmetic Task", "Feedback"

            If a study part has no division into finer phases (or a phase has no division into finer subphases) the
            dictionary value can be set to ``None``. If the whole study has no division into different parts, the
            ``structure`` dict can be set to ``None``. Default: ``None``
        test_times : list, optional
            start and end time of psychological test (in minutes) or ``None`` if the protocol has no particular test.
            ``test_times`` is then internally set to [0, 0]. Default: ``None``
        **kwargs
            additional parameters to be passed to ``BaseProtocol``, such as:

            * ``saliva_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.saliva_plot`
            * ``hr_mean_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_mean_plot`
            * ``hr_ensemble_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`


        Examples
        --------
        >>> from biopsykit.protocols import BaseProtocol
        >>> # Example 1: study with three parts, no finer division into phases
        >>> structure = {
        >>>     "Preface": None,
        >>>     "Test": None,
        >>>     "Questionnaires": None
        >>> }
        >>> BaseProtocol(name="Base", structure=structure)
        >>> # Example 2: study with three parts, all parts have different phases with specific durations
        >>> structure = {
        >>>     "Preface": {"Questionnaires": 240, "Baseline": 60},
        >>>     "Test": {"Preparation": 120, "Test": 240, "Recovery": 120},
        >>>     "Recovery": {"Part1": 240, "Part2": 240}
        >>> }
        >>> BaseProtocol(name="Base", structure=structure)
        >>> # Example 3: only certain study parts have different phases (example: TSST)
        >>> structure = {
        >>>     "Before": None,
        >>>     "TSST": {"Preparation": 300, "Talk": 300, "Math": 300},
        >>>     "After": None
        >>> }
        >>> BaseProtocol(name="Base", structure=structure)
        >>> # Example 4: study with phases and subphases, only certain study parts have different phases (example: MIST)
        >>> structure = {
        >>>     "Before": None,
        >>>     "MIST": {
        >>>         "MIST1": {"BL": 60, "AT": 240, "FB": 120},
        >>>         "MIST2": {"BL": 60, "AT": 240, "FB": 120},
        >>>         "MIST3": {"BL": 60, "AT": 240, "FB": 120}
        >>>     },
        >>>     "After": None
        >>> }
        >>> BaseProtocol(name="Base", structure=structure)

        """
        self.name: str = name
        """Study or protocol name"""

        self.structure: Dict[str, Any] = structure
        """Structure of protocol, i.e., whether protocol is divided into different parts, phases, or subphases.

        If protocol is not divided into different parts ``protocol_structure`` is set to ``None``.
        """

        self.saliva_types: Sequence[str] = []
        """List of saliva data types present in the study."""

        if test_times is None:
            test_times = [0, 0]
        self.test_times: Sequence[int] = test_times
        """Start and end time of psychological test (in minutes).

        If no psychological test was performed in the protocol ``test_times`` is set to [0, 0].
        """

        self.sample_times: Dict[str, Sequence[int]] = {}
        """Dictionary with sample times of saliva samples (in minutes).

        Sample times are either provided explicitly using the ``sample_times`` parameter in
        :meth:`~biopsykit.protocols.BaseProtocol.add_saliva_data` or by extracting it from the saliva data
        (if a ``time`` column is present).
        """

        self.saliva_data: Dict[str, SalivaRawDataFrame] = {}
        """Dictionary with saliva data collected during the study.

        Data in :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame` format can be added using
        :meth:`~biopsykit.protocols.BaseProtocol.add_saliva_data`.
        """

        self.hr_data: Dict[str, HeartRateSubjectDataDict] = {}
        """Dictionary with heart rate data collected during the study.
        If the study consists of multiple study parts each part has its own ``HeartRateSubjectDataDict``.
        If the study has no individual study parts (only different phases), the name of the one and only study part
        defaults to ``Study`` (to ensure consistent dictionary structure).

        Data in :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict` format can be added using
        :meth:`~biopsykit.protocols.BaseProtocol.add_hr_data`.
        """

        self.rpeak_data: Dict[str, SubjectDataDict] = {}
        """Dictionary with R peak data collected during the study.
        If the study consists of multiple study parts each part has its own ``SubjectDataDict``.
        If the study has no individual study parts (only different phases), the name of the one and only study part
        defaults to ``Study`` (to ensure consistent dictionary structure).

        Data in :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict` format can be added using
        :meth:`~biopsykit.protocols.BaseProtocol.add_hr_data`.
        """

        self.hr_results: Dict[str, pd.DataFrame] = {}
        """Dictionary with heart rate results.

        Dict keys are the identifiers that are specified when computing results from ``hr_data`` using
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_results`.
        """

        self.hr_above_baseline_results: Dict[str, pd.DataFrame] = {}
        """Dictionary with heart rate above baseline results.

        Dict keys are the identifiers that are specified when computing results from ``hr_data`` using
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_above_baseline`.
        """

        self.hrv_results: Dict[str, pd.DataFrame] = {}
        """Dictionary with heart rate variability ensemble.

        Dict keys are the identifiers that are specified when computing ensemble from ``rpeak_data`` using
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hrv_results`.
        """

        self.hrv_above_baseline_results: Dict[str, pd.DataFrame] = {}
        """Dictionary with heart rate variability above baseline results.

        Dict keys are the identifiers that are specified when computing results from ``rpeak_data`` using
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hrv_above_baseline`.
        """

        self.hr_ensemble: Dict[str, Dict[str, pd.DataFrame]] = {}
        """Dictionary with merged heart rate data for heart rate ensemble plot.


        Dict keys are the identifiers that are specified when computing ensemble HR data from ``hr_data`` using
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`.

        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`
            heart rate ensemble plot
        """

        self.saliva_plot_params: Dict[str, Any] = kwargs.get("saliva_plot_params", {})
        """Plot parameters for customizing the general `saliva plot` for a specific psychological protocol.

        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.saliva_plot`
            saliva plot
        """

        self.hr_mean_plot_params: Dict[str, Any] = kwargs.get("hr_mean_plot_params", {})
        """Plot parameters for customizing the general `HR mean plot` for a specific psychological protocol.

        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.hr_mean_plot`
            HR mean plot
        """

        self.hr_ensemble_plot_params: Dict[str, Any] = kwargs.get("hr_ensemble_plot_params", {})
        """Plot parameters for customizing the general `HR ensemble plot` for a specific psychological protocol.

        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`
            HR ensemble plot
        """

    def __repr__(self) -> str:
        """Return string representation of Protocol instance.

        Returns
        -------
        str
            string representation of Protocol instance

        """
        return self.__str__()

    def __str__(self) -> str:
        """Return string representation of Protocol instance.

        Returns
        -------
        str
            string representation of Protocol instance

        """
        if len(self.saliva_data) > 0:
            return """{}
            Saliva Type(s): {}
            Saliva Sample Times: {}
            Structure: {}
            """.format(
                self.name, self.saliva_types, self.sample_times, self.structure
            )
        return """{}
        Structure: {}""".format(
            self.name, self.structure
        )

    def to_file(self, file_path: path_t):
        """Serialize ``Protocol`` object and export as file.

        This function converts the basic information of this object (``name``, ``structure``, ``test_times``)
        to a JSON object and saves the serialized object to a JSON file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to export

        """
        # ensure pathlib
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".json")

        to_export = ["name", "structure", "test_times"]
        json_dict = {key: self.__dict__[key] for key in to_export}
        with file_path.open(mode="w+", encoding="utf-8") as fp:
            json.dump(json_dict, fp)

    @classmethod
    def from_file(cls: Type[T], file_path: path_t) -> T:
        """Load serialized ``Protocol`` object from file.

        Parameters
        ----------
        file_path : :class:`~pathlib.Path` or str
            file path to export

        Returns
        -------
        instance of :class:`~biopsykit.protocols.BaseProtocol`
            ``Protocol`` instance

        """
        file_path = Path(file_path)
        _assert_file_extension(file_path, ".json")

        with file_path.open(encoding="utf-8") as fp:
            json_dict = json.load(fp)
            return cls(**json_dict)

    def add_saliva_data(
        self,
        saliva_data: Union[SalivaRawDataFrame, Dict[str, SalivaRawDataFrame]],
        saliva_type: Optional[Union[str, Sequence[str]]] = None,
        sample_times: Optional[Union[Sequence[int], Dict[str, Sequence[int]]]] = None,
        test_times: Optional[Sequence[int]] = None,
    ):
        """Add saliva data collected during psychological protocol to ``Protocol`` instance.

        Parameters
        ----------
        saliva_data : :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame` or dict
            saliva data (or dict of such) to be added to this protocol.
        saliva_type : str or list of str, optional
            saliva type (or list of such) of saliva data. Not needed if ``saliva_data`` is a dictionary, then the
            saliva types are inferred from the dictionary keys.
        sample_times : list of int or dict, optional
            list of sample times in minutes. Sample times are expected to be provided *relative* to the psychological
            test in the protocol (if present). Per convention, a sample collected **directly before** was collected at
            time point :math:`t = -1`, a sample collected **directly after** the test was collected at time point
            :math`t = 0`.
        test_times : list of int, optional
            list with start and end time of psychological test in minutes. Per convention, the start of the test
            should be at time point :math:`t = 0`. ``test_times`` is also used to compute the **absolute** sample times

        """
        saliva_data = deepcopy(saliva_data)
        if isinstance(saliva_data, dict):
            saliva_type = list(saliva_data.keys())
        if isinstance(saliva_type, str):
            saliva_type = [saliva_type]

        if test_times is not None:
            self.test_times = test_times

        if saliva_data is not None:
            if not isinstance(sample_times, dict):
                sample_times = {key: sample_times for key in saliva_type}
            if not isinstance(saliva_data, dict):
                saliva_data = {key: saliva_data for key in saliva_type}
            self.sample_times.update(_get_sample_times(saliva_data, sample_times, self.test_times))
            self.saliva_data.update(self._add_saliva_data(saliva_data, saliva_type, self.sample_times))
            self.saliva_types = list(self.saliva_data.keys())

    def _add_saliva_data(
        self,
        data: Union[SalivaRawDataFrame, Dict[str, SalivaRawDataFrame]],
        saliva_type: Union[str, Sequence[str]],
        sample_times: Union[Sequence[int], Dict[str, Sequence[int]]],
    ) -> Union[SalivaRawDataFrame, Dict[str, SalivaRawDataFrame]]:
        saliva_data = {}
        if isinstance(data, dict):
            for key, value in data.items():
                saliva_data[key] = self._add_saliva_data(value, key, sample_times[key])
            return saliva_data
        is_raw = is_saliva_raw_dataframe(data, saliva_type, raise_exception=False)
        is_mse = is_saliva_mean_se_dataframe(data, raise_exception=False)
        if not any([is_mse, is_raw]):
            try:
                is_saliva_raw_dataframe(data, saliva_type)
                is_saliva_mean_se_dataframe(data)
            except ValidationError as e:
                raise ValidationError(
                    "'data' is expected to be either a SalivaRawDataFrame or a SalivaMeanSeDataFrame! "
                    "The validation raised the following error:\n\n{}".format(str(e))
                ) from e
        _check_sample_times_match(data, sample_times)
        return data

    def add_hr_data(
        self,
        hr_data: HeartRateSubjectDataDict,
        rpeak_data: Optional[SubjectDataDict] = None,
        study_part: Optional[str] = None,
    ):
        """Add time-series heart rate data collected during psychological protocol to ``Protocol`` instance.

        Parameters
        ----------
        hr_data : :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`
            dictionary with heart rate data of all subjects collected during the protocol.
        rpeak_data : :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`, optional
            dictionary with rpeak data of all subjects collected during the protocol. Needed if heart rate
            variability should be computed.
        study_part : str, optional
            string indicating to which study part data belongs to or ``None`` if data has no individual study parts.
            Default: ``None``

        """
        is_hr_subject_data_dict(hr_data)
        if study_part is None:
            study_part = "Study"
        self.hr_data[study_part] = deepcopy(hr_data)
        if rpeak_data is not None:
            rpeak_data = deepcopy(rpeak_data)
            is_subject_data_dict(rpeak_data)
            self.rpeak_data[study_part] = rpeak_data

    def compute_hr_results(  # pylint:disable=too-many-branches # noqa: C901
        self,
        result_id: str,
        study_part: Optional[str] = None,
        resample_sec: Optional[bool] = True,
        normalize_to: Optional[bool] = False,
        select_phases: Optional[bool] = False,
        split_into_subphases: Optional[bool] = False,
        mean_per_subject: Optional[bool] = True,
        add_conditions: Optional[bool] = False,
        reindex: Optional[bool] = False,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Compute heart rate data results from one study part.

        The different processing steps can be enabled or disabled by setting the function parameters to
        ``True`` or ``False``, respectively. Parameters that are required for a specific processing step can be
        provided in the ``params`` dict. The dict key must match the name of the processing step.

        Parameters
        ----------
        result_id : str
            Result ID, a descriptive name of the results that were computed.
            This ID will also be used as key to store the computed results in the ``hr_results`` dictionary.
        study_part : str, optional
            study part the data which should be processed belongs to or ``None`` if data has no
            individual study parts.
            Default: ``None``
        resample_sec : bool, optional
            ``True`` to apply resampling. Instantaneous heart rate data will then be resampled to 1 Hz.
            Default: ``True``
        normalize_to : bool, optional
            ``True`` to normalize heart rate data per subject. Data will then be the heart rate increase relative to
            the average heart rate in the phase. The name of the phase (or a dataframe containing heart rate
            data to normalize to) is specified in the ``params`` dictionary (key: ``normalize_to``).
            Default: ``False``
        select_phases : bool, optional
            ``True`` to only select specific phases for further processing, ``False`` to use all data from
            ``study_part``. The phases to be selected are specified in the ``params`` dictionary
            (key: ``select_phases``).
            Default: ``False``
        split_into_subphases : bool, optional
            ``True`` to further split phases into subphases, ``False`` otherwise. The subphases are provided as
            dictionary (keys: subphase names, values: subphase durations in seconds) in the ``params`` dictionary
            (key: ``split_into_subphases``).
            Default: ``False``
        mean_per_subject : bool, optional
            ``True`` to compute the mean heart rate per phase (and subphase, if present) for each subject and
            combine results into one dataframe. The resulting index level names of the dataframe can be provided as list
            in the ``params`` dictionary (key: ``mean_per_subject``). By default, the index level names are
            ["subject", "phase"].
            Default: ``True``
        add_conditions : bool, optional
            ``True`` to add subject conditions to dataframe data. Information on which subject belongs to which
            condition can be provided as :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
            :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` in the ``params`` dictionary
            (key: ``add_conditions``).
            Default: ``False``
        reindex : bool, optional
            ``True`` to reindex index levels of the resulting dataframe. Reindex index levels can be provided
            in the ``params`` dictionary as a dictionary with index levels as keys and index orders as values
            (key: ``reindex``).
            Default: ``False``
        params : dict, optional
            dictionary with parameters provided to the different processing steps

        """
        if study_part is None:
            study_part = "Study"
        data_dict = self.hr_data[study_part].copy()

        if params is None:
            params = {}

        if resample_sec:
            data_dict = resample_dict_sec(data_dict)

        if normalize_to:
            param = params.get("normalize_to", None)
            data_dict = normalize_to_phase(data_dict, param)

        if select_phases:
            param = params.get("select_phases", None)
            data_dict = select_dict_phases(data_dict, param)

        if split_into_subphases:
            param = params.get("split_into_subphases", None)
            data_dict = split_dict_into_subphases(data_dict, param)

        if mean_per_subject:
            if split_into_subphases:
                param = params.get("mean_per_subject", ["subject", "phase", "subphase"])
            else:
                param = params.get("mean_per_subject", ["subject", "phase"])
            data_dict = mean_per_subject_dict(data_dict, param, "Heart_Rate")

        if add_conditions:
            param = params.get("add_conditions", None)
            data_dict = add_subject_conditions(data_dict, param)

        if reindex:
            param = params.get("reindex", None)
            data_dict = BaseProtocol._reindex_df(data_dict, param)

        self.hr_results[result_id] = data_dict

    @staticmethod
    def _reindex_df(data: pd.DataFrame, param: Dict[str, Any]) -> pd.DataFrame:
        if not isinstance(param, dict):
            raise TypeError(
                "If 'reindex' is 'True', a dictionary with dataframe levels as keys "
                "and new index orders as values is expected!"
            )
        for level, index_values in param.items():
            data = data.reindex(index_values, level=level)
        return data

    def compute_hrv_results(  # pylint:disable=too-many-branches
        self,
        result_id: str,
        study_part: Optional[str] = None,
        select_phases: Optional[bool] = False,
        split_into_subphases: Optional[bool] = False,
        dict_levels: Sequence[str] = None,
        hrv_params: Optional[Dict[str, Any]] = None,
        add_conditions: Optional[bool] = False,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Compute heart rate variability ensemble from one study part.

        The different processing steps can be enabled or disabled by setting the function parameters to
        ``True`` or ``False``, respectively. Parameters that are required for a specific processing step can be
        provided in the ``params`` dict. The dict key must match the name of the processing step.

        Parameters
        ----------
        result_id : str
            Result ID, a descriptive name of the ensemble that were computed.
            This ID will also be used as key to store the computed ensemble in the ``hrv_results`` dictionary.
        study_part : str, optional
            study part the data which should be processed belongs to or ``None`` if data has no
            individual study parts.
            Default: ``None``
        select_phases : bool, optional
            ``True`` to only select specific phases for further processing, ``False`` to use all data from
            ``study_part``. The phases to be selected are specified in the ``params`` dictionary
            (key: ``select_phases``).
            Default: ``False``
        split_into_subphases : bool, optional
            ``True`` to further split phases into subphases, ``False`` otherwise. The subphases are provided as
            dictionary (keys: subphase names, values: subphase durations in seconds) in the ``params`` dictionary
            (key: ``split_into_subphases``).
            Default: ``False``
        dict_levels : list, optional
            list with names of dictionary levels which will also be the index level names of the resulting dataframe
            or ``None`` to use default level names: ["subject", "phase"] (if ``split_into_subphases`` is ``False``)
            or ["subject", "phase", "subphase"] (if ``split_into_subphases`` is ``True``).
        hrv_params : dict, optional
            dictionary with parameters to configure HRV processing or ``None`` to use default parameter.
            See :func:`~biopsykit.signals.ecg.EcgProcessor.hrv_process` for an overview on available parameters.
        add_conditions : bool, optional
            ``True`` to add subject conditions to dataframe data. Information on which subject belongs to which
            condition can be provided as :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
            :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` in the ``params`` dictionary
            (key: ``add_conditions``).
            Default: ``False``
        params : dict, optional
            dictionary with parameters provided to the different processing steps.

        """
        if study_part is None:
            study_part = "Study"
        if dict_levels is None:
            dict_levels = ["subject", "phase"]
            if split_into_subphases:
                dict_levels.append("subphase")

        data_dict = self.rpeak_data[study_part].copy()

        if hrv_params is None:
            hrv_params = {}
        if params is None:
            params = {}

        if select_phases:
            param = params.get("select_phases", None)
            data_dict = select_dict_phases(data_dict, param)

        if split_into_subphases:
            param = params.get("split_into_subphases", None)
            data_dict = split_dict_into_subphases(data_dict, param)

        hrv_result = self._compute_hrv_dict(data_dict, hrv_params, dict_levels)
        # drop most inner level (comes from neurokit's hrv function and is not needed)
        hrv_result = hrv_result.droplevel(level=-1)

        if add_conditions:
            param = params.get("add_conditions", None)
            hrv_result = add_subject_conditions(hrv_result, param)
        self.hrv_results[result_id] = hrv_result

    def _compute_hrv_dict(
        self, rpeak_dict: Dict[str, Any], hrv_params: Dict[str, Any], dict_levels: Sequence[str]
    ) -> pd.DataFrame:
        result_dict = {}
        for key, value in tqdm(list(rpeak_dict.items()), desc=dict_levels[0]):
            _assert_is_dtype(value, (dict, pd.DataFrame))
            if isinstance(value, dict):
                # nested dictionary
                result_dict[key] = self._compute_hrv_dict(value, hrv_params, dict_levels[1:])
            else:
                result_dict[key] = EcgProcessor.hrv_process(rpeaks=value, **hrv_params)

        return pd.concat(result_dict, names=[dict_levels[0]])

    def compute_hr_ensemble(  # pylint:disable=too-many-branches
        self,
        ensemble_id: str,
        study_part: Optional[str] = None,
        resample_sec: Optional[bool] = True,
        normalize_to: Optional[bool] = True,
        select_phases: Optional[bool] = False,
        cut_phases: Optional[bool] = True,
        merge_dict: Optional[bool] = True,
        add_conditions: Optional[bool] = False,
        params: Dict[str, Any] = None,
    ):
        """Compute heart rate ensemble data from one study part.

        Heart rate ensemble data are time-series data where data from all subjects within one phase
        have the same length and can thus be overlaid as mean ± standard error in a plot.

        The different processing steps can be enabled or disabled by setting the function parameters to
        ``True`` or ``False``, respectively. Parameters that are required for a specific processing step can be
        provided in the ``params`` dict. The dict key must match the name of the processing step.


        Parameters
        ----------
        ensemble_id : str
            ensemble identifier, a descriptive name of the ensemble data that were computed.
            This ID will also be used as key to store the computed ensemble data in the ``hr_ensemble`` dictionary.
        study_part : str, optional
            study part the data which should be processed belongs to or ``None`` if data has no
            individual study parts.
            Default: ``None``
        resample_sec : bool, optional
            ``True`` to apply resampling. Instantaneous heart rate data will then be resampled to 1 Hz.
            Default: ``True``
        normalize_to : bool, optional
            ``True`` to normalize heart rate data per subject. Data will then be the heart rate increase relative to
            the average heart rate in the phase. The name of the phase (or a dataframe containing heart rate
            data to normalize to) is specified in the ``params`` dictionary (key: ``normalize_to``).
            Default: ``False``
        select_phases : bool, optional
            ``True`` to only select specific phases for further processing, ``False`` to use all data from
            ``study_part``. The phases to be selected are specified in the ``params`` dictionary
            (key: ``select_phases``).
            Default: ``False``
        cut_phases : bool, optional
            ``True`` to cut time-series data to shortest duration of a subject in each phase, ``False`` otherwise.
            Default: ``True``
        merge_dict : bool, optional
            ``True`` to convert :obj:`~biopsykit.utils.datatype_helper.StudyDataDict` into
            :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`, i.e., merge dictionary data from
            individual subjects into one dataframe for each phase.
            Default: ``True``
        add_conditions : bool, optional
            ``True`` to add subject conditions to dataframe data. Information on which subject belongs to which
            condition can be provided as :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
            :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` in the ``params`` dictionary
            (key: ``add_conditions``).
            Default: ``False``
        params : dict, optional
            dictionary with parameters provided to the different processing steps

        See Also
        --------
        :func:`~biopsykit.protocols.plotting.hr_ensemble_plot`
            Heart rate ensemble plot

        """
        if study_part is None:
            study_part = "Study"
        data_dict = self.hr_data[study_part].copy()

        if params is None:
            params = {}

        if resample_sec:
            data_dict = resample_dict_sec(data_dict)

        if normalize_to:
            param = params.get("normalize_to", None)
            if param is None:
                raise ValueError("When 'normalize_to' is 'True' a phase name must be specified!")
            data_dict = normalize_to_phase(data_dict, param)

        data_dict = rearrange_subject_data_dict(data_dict)

        if select_phases:
            param = params.get("select_phases", data_dict.keys())
            data_dict = {phase: data_dict[phase] for phase in param}

        if cut_phases:
            data_dict = cut_phases_to_shortest(data_dict)

        if merge_dict:
            param = params.get("merge_dict", None)
            data_dict = merge_study_data_dict(data_dict, param)

        if add_conditions:
            param = params.get("add_conditions", None)
            data_dict = split_subject_conditions(data_dict, param)

        self.hr_ensemble[ensemble_id] = data_dict

    def compute_hr_above_baseline(  # pylint:disable=too-many-branches
        self,
        result_id: str,
        baseline_phase: str,
        study_part: Optional[str] = None,
        select_phases: Optional[bool] = False,
        split_into_subphases: Optional[bool] = False,
        add_conditions: Optional[bool] = False,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Compute the relative amount of heart rate above a specified baseline.

        The different processing steps can be enabled or disabled by setting the function parameters to
        ``True`` or ``False``, respectively. Parameters that are required for a specific processing step can be
        provided in the ``params`` dict. The dict key must match the name of the processing step.

        Parameters
        ----------
        result_id : str
            Result ID, a descriptive name of the results that were computed.
            This ID will also be used as key to store the computed results in the ``hr_above_baseline_results``
            dictionary.
        baseline_phase : str
            string indicating the name of the phase that should be used as baseline for computing the relative amount
            above the baseline.
        study_part : str, optional
            study part the data which should be processed belongs to or ``None`` if data has no
            individual study parts.
            Default: ``None``
        select_phases : bool, optional
            ``True`` to only select specific phases for further processing, ``False`` to use all data from
            ``study_part``. The phases to be selected are specified in the ``params`` dictionary
            (key: ``select_phases``).
            Default: ``False``
        split_into_subphases : bool, optional
            ``True`` to further split phases into subphases, ``False`` otherwise. The subphases are provided as
            dictionary (keys: subphase names, values: subphase durations in seconds) in the ``params`` dictionary
            (key: ``split_into_subphases``).
            Default: ``False``
        add_conditions : bool, optional
            ``True`` to add subject conditions to dataframe data. Information on which subject belongs to which
            condition can be provided as :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
            :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` in the ``params`` dictionary
            (key: ``add_conditions``).
            Default: ``False``
        params : dict, optional
            dictionary with parameters provided to the different processing steps

        """
        if study_part is None:
            study_part = "Study"
        data_dict = self.hr_data[study_part].copy()

        if params is None:
            params = {}

        data_dict = normalize_to_phase(data_dict, baseline_phase)

        if select_phases:
            param = params.get("select_phases", None)
            data_dict = select_dict_phases(data_dict, param)

        if split_into_subphases:
            param = params.get("split_into_subphases", None)
            data_dict = split_dict_into_subphases(data_dict, param)

        data_dict = BaseProtocol._compute_hr_above_baseline(data_dict)

        result_data = pd.concat(data_dict, names=["subject"]).unstack()

        if add_conditions:
            param = params.get("add_conditions")
            result_data = result_data.join(param.set_index("condition", append=True))
            index_levels = result_data.index.names
            result_data = result_data.reorder_levels([index_levels[-1]] + index_levels[:-1])

        self.hr_above_baseline_results[result_id] = result_data

    def compute_hrv_above_baseline(  # pylint:disable=too-many-branches
        self,
        result_id: str,
        baseline_phase: str,
        continuous_hrv_data: SubjectDataDict,
        select_phases: Optional[bool] = False,
        split_into_subphases: Optional[bool] = False,
        add_conditions: Optional[bool] = False,
        hrv_columns: Optional[Sequence[str]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Compute the relative amount of heart rate variability above a specified baseline.

        The different processing steps can be enabled or disabled by setting the function parameters to
        ``True`` or ``False``, respectively. Parameters that are required for a specific processing step can be
        provided in the ``params`` dict. The dict key must match the name of the processing step.

        Parameters
        ----------
        result_id : str
            Result ID, a descriptive name of the results that were computed.
            This ID will also be used as key to store the computed results in the ``hrv_above_baseline_results``
            dictionary.
        baseline_phase : str
            string indicating the name of the phase that should be used as baseline for computing the relative amount
            above the baseline.
        continuous_hrv_data : :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`
            dictionary with continuous HRV of all subjects collected during the protocol.
        select_phases : bool, optional
            ``True`` to only select specific phases for further processing, ``False`` to use all data from
            ``study_part``. The phases to be selected are specified in the ``params`` dictionary
            (key: ``select_phases``).
            Default: ``False``
        split_into_subphases : bool, optional
            ``True`` to further split phases into subphases, ``False`` otherwise. The subphases are provided as
            dictionary (keys: subphase names, values: subphase durations in seconds) in the ``params`` dictionary
            (key: ``split_into_subphases``).
            Default: ``False``
        add_conditions : bool, optional
            ``True`` to add subject conditions to dataframe data. Information on which subject belongs to which
            condition can be provided as :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` or
            :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` in the ``params`` dictionary
            (key: ``add_conditions``).
            Default: ``False``
        hrv_columns: list of str
            selected column names for computing or ``None`` to compute relative amount above baseline for all columns.
            Default: ``None``
        params : dict, optional
            dictionary with parameters provided to the different processing steps

        """
        data_dict = deepcopy(continuous_hrv_data)

        if params is None:
            params = {}

        data_dict_baseline = select_dict_phases(data_dict, baseline_phase)

        if select_phases:
            param = params.get("select_phases", None)
            data_dict = select_dict_phases(data_dict, param)

        if split_into_subphases:
            param = params.get("split_into_subphases", None)
            data_dict = split_dict_into_subphases(data_dict, param)

        data_dict = BaseProtocol._compute_hrv_above_baseline(data_dict, data_dict_baseline, hrv_columns)

        result_data = pd.concat(data_dict, names=["subject"]).unstack()

        if add_conditions:
            param = params.get("add_conditions")
            result_data = result_data.join(param.set_index("condition", append=True))
            index_levels = result_data.index.names
            result_data = result_data.reorder_levels([index_levels[-1]] + index_levels[:-1])

        self.hrv_above_baseline_results[result_id] = result_data

    @staticmethod
    def _compute_hr_above_baseline(data_dict: Dict[str, Union[Dict[str, pd.DataFrame], pd.DataFrame]]):
        result_dict = {}
        for subject_id, hr_data_dict in data_dict.items():
            dict_hr = {}
            for phase, dict_hr_phase in hr_data_dict.items():
                if isinstance(dict_hr_phase, dict):
                    dict_hr_subphase = {}
                    for subphase, df_hr in dict_hr_phase.items():
                        hr_above_bl = (df_hr > 0).sum() / len(df_hr) * 100
                        dict_hr_subphase[subphase] = hr_above_bl
                    dict_hr[phase] = pd.concat(dict_hr_subphase, names=["subphase"])
                else:
                    hr_above_bl = (dict_hr_phase > 0).sum() / len(dict_hr_phase) * 100
                    dict_hr[phase] = hr_above_bl

            result_dict[subject_id] = pd.concat(dict_hr, names=["phase"])
        return result_dict

    @staticmethod
    def _compute_hrv_above_baseline(
        data_dict: Dict[str, Union[Dict[str, pd.DataFrame], pd.DataFrame]],
        data_dict_baseline: Dict[str, Union[Dict[str, pd.DataFrame], pd.DataFrame]],
        hrv_columns: Optional[Sequence[str]] = None,
    ):
        result_dict = {}
        for subject_id, hrv_data_dict in data_dict.items():
            dict_hrv = {}
            hrv_baseline = data_dict_baseline[subject_id]
            hrv_baseline = list(hrv_baseline.values())[0]
            if hrv_columns is None:
                hrv_columns = hrv_baseline.columns
            hrv_baseline = hrv_baseline[hrv_columns]

            for phase, dict_hrv_phase in hrv_data_dict.items():
                if isinstance(dict_hrv_phase, dict):
                    dict_hrv_subphases = {}
                    for subphase, df_hrv in dict_hrv_phase.items():
                        df_hrv_slice = df_hrv[hrv_columns]
                        hrv_above_bl = (df_hrv_slice > hrv_baseline.mean()).sum() / len(df_hrv_slice) * 100
                        dict_hrv_subphases[subphase] = hrv_above_bl
                    dict_hrv[phase] = pd.concat(dict_hrv_subphases, names=["subphase"])
                else:
                    hrv_above_bl = (dict_hrv_phase > hrv_baseline.mean()).sum() / len(dict_hrv_phase) * 100
                    dict_hrv[phase] = hrv_above_bl

            result_dict[subject_id] = pd.concat(dict_hrv, names=["phase"])
        return result_dict

    def add_hr_results(self, result_id: str, results: pd.DataFrame):
        """Add existing heart rate processing ensemble.

        Parameters
        ----------
        result_id : str
            identifier of result parameters used to store dataframe in ``hr_results`` dictionary.
        results : :class:`~pandas.DataFrame`
            dataframe with computed heart rate processing ensemble

        """
        _assert_is_dtype(results, pd.DataFrame)
        results = deepcopy(results)
        self.hr_results[result_id] = results

    def get_hr_results(self, result_id: str) -> pd.DataFrame:
        """Return heart rate processing results.

        Heart rate results can be computed by calling :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_results`.

        Parameters
        ----------
        result_id : str
            identifier of result parameters specified when computing results via
            :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_results`

        Returns
        -------
        :class:`~pandas.DataFrame`
            heart rate processing results

        """
        return self.hr_results.get(result_id, None)

    def add_hrv_results(self, result_id: str, results: pd.DataFrame):
        """Add existing heart rate variability processing ensemble.

        Parameters
        ----------
        result_id : str
            identifier of result parameters used to store dataframe in ``hrv_results`` dictionary
        results : :class:`~pandas.DataFrame`
            dataframe with computed heart rate variability processing ensemble

        """
        results = deepcopy(results)
        self.hrv_results[result_id] = results

    def export_hr_results(self, base_path: path_t, prefix: Optional[str] = None):
        """Export all heart rate results to csv files.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            folder path to export all heart rate result files to
        prefix : str, optional
            prefix to add to file name or ``None`` to use ``name`` attribute (in lowercase) as prefix

        """
        self._export_results(base_path, prefix, self.hr_results)

    def export_hr_above_baseline_results(self, base_path: path_t, prefix: Optional[str] = None):
        """Export all heart rate above baseline results to csv files.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            folder path to export all heart rate above baseline result files to
        prefix : str, optional
            prefix to add to file name or ``None`` to use ``name`` attribute (in lowercase) as prefix

        """
        self._export_results(base_path, prefix, self.hr_above_baseline_results)

    def export_hr_ensemble(self, base_path: path_t, prefix: Optional[str] = None):
        """Export all heart rate ensemble data to Excel files.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            folder path to export all heart rate ensemble files to
        prefix : str, optional
            prefix to add to file name or ``None`` to use ``name`` attribute (in lowercase) as prefix

        """
        self._export_ensemble(base_path, prefix, self.hr_ensemble)

    def export_hrv_results(self, base_path: path_t, prefix: Optional[str] = None):
        """Export all heart rate variability results to csv files.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            folder path to export all heart rate variability result files to
        prefix : str, optional
            prefix to add to file name or ``None`` to use ``name`` attribute (in lowercase) as prefix

        """
        self._export_results(base_path, prefix, self.hrv_results)

    def export_hrv_above_baseline_results(self, base_path: path_t, prefix: Optional[str] = None):
        """Export all heart rate variability above baseline results to csv files.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            folder path to export all heart rate variability above baseline result files to
        prefix : str, optional
            prefix to add to file name or ``None`` to use ``name`` attribute (in lowercase) as prefix

        """
        self._export_results(base_path, prefix, self.hrv_above_baseline_results)

    def _export_results(self, base_path: path_t, prefix: str, result_dict: Dict[str, pd.DataFrame]):
        # ensure pathlib
        base_path = Path(base_path)
        if not base_path.is_dir():
            raise ValueError("'base_path' must be a directory!")
        if prefix is None:
            prefix = self.name.lower().replace(" ", "_")
        for key, data in result_dict.items():
            file_name = f"{prefix}_{key}.csv"
            data.to_csv(base_path.joinpath(file_name))

    def _export_ensemble(self, base_path: path_t, prefix: str, result_dict: Dict[str, Dict[str, pd.DataFrame]]):
        # ensure pathlib
        base_path = Path(base_path)
        if not base_path.is_dir():
            raise ValueError("'base_path' must be a directory!")
        if prefix is None:
            prefix = self.name.lower().replace(" ", "_")
        for key, data in result_dict.items():
            file_name = f"{prefix}_{key}.xlsx"
            write_pandas_dict_excel(data, base_path.joinpath(file_name))

    def get_hrv_results(self, result_id: str) -> pd.DataFrame:
        """Return heart rate variability processing ensemble.

        Heart rate variability ensemble can be computed by calling
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hrv_results`.

        Parameters
        ----------
        result_id : str
            identifier of result parameters specified when computing ensemble via
            :meth:`~biopsykit.protocols.BaseProtocol.compute_hrv_results`

        Returns
        -------
        :class:`~pandas.DataFrame`
            heart rate variability processing ensemble

        """
        return self.hrv_results.get(result_id, None)

    def add_hr_ensemble(self, ensemble_id: str, ensemble: Dict[str, pd.DataFrame]):
        """Add existing heart rate ensemble data.

        Parameters
        ----------
        ensemble_id : str
            identifier of ensemble parameters used to store dictionary in ``hr_ensemble`` dictionary
        ensemble : :class:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`
            ensemble data as ``MergedStudyDataDict``

        """
        _assert_is_dtype(ensemble, dict)
        ensemble = deepcopy(ensemble)
        self.hr_ensemble[ensemble_id] = ensemble

    def get_hr_ensemble(self, ensemble_id: str):
        """Return heart rate ensemble data.

        Parameters
        ----------
        ensemble_id : str
            identifier of ensemble parameters specified when computing ensemble parameters via
            :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`

        Returns
        -------
        :class:`~pandas.DataFrame`
            heart rate ensemble ensemble

        """
        return self.hr_ensemble.get(ensemble_id, None)

    def saliva_plot(
        self,
        saliva_type: Optional[Union[str, Sequence[str]]] = "cortisol",
        **kwargs,
    ) -> Optional[Tuple[plt.Figure, plt.Axes]]:
        """Plot saliva data during psychological protocol as mean ± standard error.

        Parameters
        ----------
        saliva_type : {"cortisol", "amylase", "il6"}, optional
            saliva type to be plotted. If a dict is passed and ``saliva_type`` is ``None``
            the saliva types are inferred from dict keys. Default: ``cortisol``
        **kwargs
            additional parameters to be passed to :func:`~biopsykit.protocols.plotting.saliva_plot`.


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        ax : :class:`matplotlib.axes.Axes`
            axes object

        See Also
        --------
        :func:`~biopsykit.protocols.plotting.saliva_plot`
            Plot saliva data during a psychological protocol

        """
        if len(self.saliva_types) == 0:
            raise ValueError("No saliva data to plot!")

        for key, val in self.saliva_plot_params.items():
            kwargs.setdefault(key, val)

        if isinstance(saliva_type, str):
            saliva_type = [saliva_type]

        data = {key: self.saliva_data[key] for key in saliva_type}
        sample_times = {key: self.sample_times[key] for key in saliva_type}

        return plot.saliva_plot(
            data=data,
            saliva_type=None,
            sample_times=sample_times,
            test_times=self.test_times,
            sample_times_absolute=True,
            **kwargs,
        )

    @staticmethod
    def saliva_plot_combine_legend(fig: plt.Figure, ax: plt.Axes, saliva_types: Sequence[str], **kwargs):
        """Combine multiple legends of :func:`~biopsykit.protocols.plotting.saliva_plot` into one legend outside plot.

        If data from multiple saliva types are combined into one plot (e.g., by calling
        :func:`~biopsykit.protocols.plotting.saliva_plot` on the same plot twice) then two separate legend are created.
        This function can be used to combine the two legends into one.


        Parameters
        ----------
        fig : :class:`~matplotlib.figure.Figure`
            figure object
        ax : :class:`~matplotlib.axes.Axes`
            axes object
        saliva_types : list
            list of saliva types in plot
        **kwargs
            additional arguments to customize plot that are passed to
            :func:`~biopsykit.protocols.plotting.saliva_plot_combine_legend`

        """
        return plot.saliva_plot_combine_legend(fig=fig, ax=ax, saliva_types=saliva_types, **kwargs)

    def saliva_feature_boxplot(
        self,
        x: str,
        saliva_type: str,
        feature: Optional[str] = None,
        stats_kwargs: Optional[Dict] = None,
        **kwargs,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Draw a boxplot with significance brackets, specifically designed for saliva features.

        This is a wrapper of :func:`~biopsykit.protocols.plotting.saliva_feature_boxplot` that directly uses the
        saliva data added to this ``Protocol`` instance.

        Parameters
        ----------
        x : str
            column of x axis in ``data``
        saliva_type : str
            type of saliva data to plot
        feature : str, optional
            name of feature to plot or ``None``
        stats_kwargs : dict, optional
            dictionary with arguments for significance brackets
        **kwargs
            additional arguments that are passed to :func:`~biopsykit.protocols.plotting.saliva_feature_boxplot`


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        ax : :class:`matplotlib.axes.Axes`
            axes object


        See Also
        --------
        :func:`~biopsykit.protocols.plotting.saliva_feature_boxplot`
            plot saliva features as boxplot without ``Protocol`` instance
        :func:`~biopsykit.plotting.feature_boxplot`
            plot features as boxplot

        """
        # TODO: add support for computing saliva features
        return plot.saliva_feature_boxplot(
            self.saliva_data[saliva_type], x, saliva_type, feature, stats_kwargs, **kwargs
        )

    @staticmethod
    def saliva_multi_feature_boxplot(
        data: SalivaFeatureDataFrame,
        saliva_type: str,
        features: Union[Sequence[str], Dict[str, Union[str, Sequence[str]]]],
        hue: Optional[str] = None,
        stats_kwargs: Optional[Dict] = None,
        **kwargs,
    ) -> Tuple[plt.Figure, Iterable[plt.Axes]]:
        """Draw multiple features as boxplots with significance brackets, specifically designed for saliva features.

        This is a wrapper of :func:`~biopsykit.protocols.plotting.saliva_multi_feature_boxplot`.

        Parameters
        ----------
        data : :class:`~biopsykit.utils.datatype_helper.SalivaFeatureDataFrame`
            saliva feature dataframe
        saliva_type : str
            type of saliva data to plot
        hue : str, optional
            column name of grouping variable. Default: ``None``
        features : list of str or dict of str
            features to plot. If ``features`` is a list, each entry must correspond to one feature category in the
            index level specified by ``group``. A separate subplot will be created for each feature.
            If similar features (i.e., different `slope` or `AUC` parameters) should be combined into one subplot,
            ``features`` can be provided as dictionary.
            Then, the dict keys specify the feature category (a separate subplot will be created for each category)
            and the dict values specify the feature (or list of features) that are combined into the subplots.
        stats_kwargs : dict, optional
            nested dictionary with arguments for significance brackets.
            See :func:`~biopsykit.plotting.feature_boxplot` for further information


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        axs : list of :class:`matplotlib.axes.Axes`
            list of subplot axes objects


        See Also
        --------
        :func:`~biopsykit.protocols.plotting.saliva_multi_feature_boxplot`
            plot multiple saliva features as boxplots without instantiating a ``Protocol`` instance
        :func:`~biopsykit.stats.StatsPipeline`
            class to create statistical analysis pipelines and get parameter for plotting significance brackets

        """
        return plot.saliva_multi_feature_boxplot(data, saliva_type, features, hue, stats_kwargs, **kwargs)

    def hr_ensemble_plot(
        self, ensemble_id: str, subphases: Optional[Dict[str, Dict[str, int]]] = None, **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        r"""Draw heart rate ensemble plot.

        Parameters
        ----------
        ensemble_id : str
            identifier of the ensemble data to be plotted.
            Ensemble data needs to be computed using :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`
            first
        subphases : dict, optional
            dictionary with phases (keys) and subphases (values - dict with subphase names and subphase durations) or
            ``None`` if no subphases are present. Default: ``None``
        **kwargs : dict, optional
            additional parameters to be passed to :func:`~biopsykit.protocols.plotting.hr_ensemble_plot` for
            plot configuration, such as:

            * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and returned.
            * ``palette``: color palette to plot data from different phases. If ``palette`` is a str then it is
              assumed to be the name of a ``fau_colors`` palette (``fau_colors.cmaps._fields``).
            * ``figsize``: tuple specifying figure dimensions
            * ``ensemble_alpha``: transparency value for ensemble plot errorband (around mean). Default: 0.3
            * ``background_alpha``: transparency value for background spans (if subphases are present). Default: 0.2
            * ``linestyle``: list of line styles for ensemble plots. Must match the number of phases to plot
            * ``phase_text``: string pattern to customize phase name shown in legend with placeholder for subphase name.
              Default: "{}"

            To style axes:

            * ``is_relative``: boolean indicating whether heart rate data is relative (in % relative to baseline)
              or absolute (in bpm). Default: ``True``
            * ``xlabel``: label of x axis. Default: ":math:`Time [s]`"
            * ``xaxis_minor_tick_locator``: locator object to style x axis minor ticks. Default: 60 sec
            * ``ylabel``: label of y axis. Default: ":math:`\Delta HR [\%]`"
            * ``ylims``: y axis limits. Default: ``None`` to automatically infer limits

            To style the annotations at the end of each phase:

            * ``end_phase_text``: string pattern to customize text at the end of phase with placeholder for phase name.
              Default: "{}"
            * ``end_phase_line_color``: line color of vertical lines used to indicate end of phase. Default: "#e0e0e0"
            * ``end_phase_line_width``: line width of vertical lines used  to indicate end of phase. Default: 2.0

            To style legend:

            * ``legend_loc``: location of legend. Default: "lower right"
            * ``legend_bbox_to_anchor``: box that is used to position the legend in conjunction with ``legend_loc``


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        ax : :class:`matplotlib.axes.Axes`
            axes object


        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`
            compute heart rate ensemble data
        :func:`~biopsykit.protocols.plotting.hr_ensemble_plot`
            Heart rate ensemble plot

        """
        data = self.hr_ensemble[ensemble_id]
        kwargs.update(self.hr_ensemble_plot_params)
        return plot.hr_ensemble_plot(data=data, subphases=subphases, **kwargs)

    def hr_mean_plot(
        self,
        result_id: str,
        **kwargs,
    ) -> Tuple[plt.Figure, plt.Axes]:
        r"""Plot course of heart rate as mean ± standard error over phases (and subphases) of a psychological protocol.

        The correct plot is automatically inferred from the provided data:

        * only ``phase`` index level: plot phases over x axis
        * ``phase`` and ``subphase`` index levels: plot subphases over x axis, highlight phases as vertical spans
        * additionally: ``condition`` level: plot data of different conditions individually
          (corresponds to ``hue`` parameter in :func:`biopsykit.plotting.lineplot`)


        Parameters
        ----------
        result_id : str
            identifier of the heart rate result data to be plotted
        **kwargs
            additional parameters to be passed to :func:`~biopsykit.protocols.plotting.hr_mean_plot` for
            plot configuration, such as:

            * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created
              and returned.
            * ``figsize``: tuple specifying figure dimensions
            * ``palette``: color palette to plot data from different conditions. If ``palette`` is a str then it is
              assumed to be the name of a ``fau_colors`` palette (``fau_colors.cmaps._fields``).
            * ``is_relative``: boolean indicating whether heart rate data is relative (in % relative to baseline)
              or absolute (in bpm). Default: ``False``
            * ``order``: list specifying the order of categorical values (i.e., conditions) along the x axis.
            * ``x_offset``: offset value to move different groups along the x axis for better visualization.
              Default: 0.05
            * ``xlabel``: label of x axis. Default: "Subphases" (if subphases are present)
              or "Phases" (if only phases are present)
            * ``ylabel``: label of y axis. Default: ":math:`\Delta HR [%]`"
            * ``ylims``: list to manually specify y axis limits, float to specify y axis margin
              (see :meth:`~matplotlib.Axes.margin()` for further information), or ``None`` to automatically infer
              y axis limits
            * ``marker``: string or list of strings to specify marker style.
              If ``marker`` is a string, then marker of each line will have the same style.
              If ``marker`` is a list, then marker of each line will have a different style.
            * ``linestyle``: string or list of strings to specify line style.
              If ``linestyle`` is a string, then each line will have the same style.
              If ``linestyle`` is a list, then each line will have a different style.


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        ax : :class:`matplotlib.axes.Axes`
            axes object


        See Also
        --------
        :func:`~biopsykit.protocols.plotting.hr_mean_plot`
            Plot heart rate data as lineplot with mean and standard error
        :func:`~biopsykit.plotting.lineplot`
            Plot generic data as lineplot with mean and standard error

        """
        data = mean_se_per_phase(self.hr_results[result_id])
        kwargs.update(self.hr_mean_plot_params)
        return plot.hr_mean_plot(data=data, **kwargs)



"""Utility functions for sleep/wake detection algorithms."""

from typing import Literal, Optional

import numpy as np

EPOCH_LENGTH = Literal[30, 60]


def rescore(predictions: np.ndarray, epoch_length: Optional[EPOCH_LENGTH] = 30) -> np.ndarray:
    """Apply Webster's rescoring rules to sleep/wake predictions.

    Parameters
    ----------
    predictions : array_like
        sleep/wake predictions
    epoch_length : int
        length of actigraphy epoch in seconds

    Returns
    -------
    array_like
        rescored sleep/wake predictions

    """
    rescored = predictions.copy()
    # rules a through c
    rescored = _apply_recording_rules_a_c(rescored, epoch_length)
    # rules d and e
    rescored = _apply_recording_rules_d_e(rescored, epoch_length)

    # wake phases of 1 minute, surrounded by sleep, get rescored
    for t in range(1, len(rescored) - 1):
        if rescored[t] == 1 and rescored[t - 1] == 0 and rescored[t + 1] == 0:
            rescored[t] = 0

    return rescored


def _apply_recording_rules_a_c(rescored: np.ndarray, epoch_length: EPOCH_LENGTH):  # pylint:disable=too-many-branches
    wake_bin = 0
    for t in range(len(rescored)):  # pylint:disable=consider-using-enumerate
        if rescored[t] == 1:
            wake_bin += 1
        elif epoch_length == 30:
            if wake_bin >= 30:
                # rule c: at least 15 minutes of wake, next 4 minutes of sleep get rescored
                rescored[t : t + 8] = 0
            elif 20 <= wake_bin < 30:
                # rule b: at least 10 minutes of wake, next 3 minutes of sleep get rescored
                rescored[t : t + 6] = 0
            elif 8 <= wake_bin < 20:
                # rule a: at least 4 minutes of wake, next 1 minute of sleep gets rescored
                rescored[t : t + 2] = 0
            wake_bin = 0
        else:
            if wake_bin >= 15:
                # rule c: at least 15 minutes of wake, next 4 minutes of sleep get rescored
                rescored[t : t + 4] = 0
            elif 10 <= wake_bin < 15:
                # rule b: at least 10 minutes of wake, next 3 minutes of sleep get rescored
                rescored[t : t + 3] = 0
            elif 4 <= wake_bin < 10:
                # rule a: at least 4 minutes of wake, next 1 minute of sleep gets rescored
                rescored[t : t + 1] = 0
            wake_bin = 0

    return rescored


def _apply_recording_rules_d_e(rescored: np.ndarray, epoch_length: EPOCH_LENGTH):  # pylint:disable=too-many-branches
    # rule d/e: 6/10 minutes or less of sleep surrounded by at least 10/20 minutes of wake on each side get rescored
    if epoch_length == 30:
        sleep_rules = [12, 20]
        wake_rules = [20, 40]
    else:
        sleep_rules = [6, 10]
        wake_rules = [10, 20]

    for sleep_thres, wake_thres in zip(sleep_rules, wake_rules):
        sleep_bin = 0
        start_ind = 0
        for t in range(wake_thres, len(rescored) - wake_thres):
            if rescored[t] == 1:
                sleep_bin += 1
                if sleep_bin == 1:
                    start_ind = t
            else:
                sum1 = np.sum(rescored[start_ind - wake_thres : start_ind])
                sum2 = np.sum(rescored[t : t + wake_thres])
                if sleep_thres >= sleep_bin > 0 == sum1 and sum2 == 0:
                    rescored[start_ind:t] = 0
                sleep_bin = 0

    return rescored



"""Module for analysis of sleep-related data, e.g. polysomnography or IMU/actigraphy data collected during sleep."""
from biopsykit.sleep import plotting, psg, sleep_endpoints, sleep_processing_pipeline, sleep_wake_detection, utils

__all__ = ["utils", "sleep_endpoints", "plotting", "psg", "sleep_wake_detection", "sleep_processing_pipeline"]



"""Module providing various functions for processing more complex structured data (e.g., collected during a study)."""
import warnings
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import (
    _assert_dataframes_same_length,
    _assert_has_index_levels,
    _assert_has_multiindex,
    _assert_is_dtype,
)
from biopsykit.utils.array_handling import sanitize_input_1d
from biopsykit.utils.datatype_helper import (
    MeanSeDataFrame,
    MergedStudyDataDict,
    StudyDataDict,
    SubjectConditionDataFrame,
    SubjectConditionDict,
    SubjectDataDict,
    _MeanSeDataFrame,
    is_mean_se_dataframe,
    is_merged_study_data_dict,
    is_study_data_dict,
    is_subject_condition_dataframe,
    is_subject_condition_dict,
    is_subject_data_dict,
)
from biopsykit.utils.functions import se
from scipy import interpolate


def _split_data_series(data: pd.DataFrame, time_intervals: pd.Series, include_start: bool) -> Dict[str, pd.DataFrame]:
    if time_intervals.index.nlevels > 1:
        # multi-index series => second level contains start/end times of phases
        time_intervals = time_intervals.unstack().T
        time_intervals = {key: tuple(value.values()) for key, value in time_intervals.to_dict().items()}
    else:
        if include_start:
            time_intervals["Start"] = data.index[0].to_pydatetime().time()
        # time_intervals.sort_values(inplace=True)
        time_intervals = {
            name: (start, end)
            for name, start, end in zip(time_intervals.index, time_intervals[:-1], time_intervals[1:])
        }
    return time_intervals


def split_data(
    data: pd.DataFrame,
    time_intervals: Union[pd.DataFrame, pd.Series, Dict[str, Sequence[str]]],
    include_start: Optional[bool] = False,
) -> Dict[str, pd.DataFrame]:
    """Split data into different phases based on time intervals.

    The start and end times of the phases are prodivded via the ``time_intervals`` parameter and can either be a
    :class:`~pandas.Series`, 1 row of a :class:`~pandas.DataFrame`, or a dictionary with start and end times per phase.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to be split
    time_intervals : dict or :class:`~pandas.Series` or :class:`~pandas.DataFrame`
        time intervals indicating where the data should be split. This can be:

        * :class:`~pandas.Series` object or 1 row of a :class:`~pandas.DataFrame` with `start` times of each phase.
          The phase names are then derived from the `index` names in case of a :class:`~pandas.Series` or from the
          `columns` names in case of a :class:`~pandas.DataFrame`.
        * dictionary with phase names (keys) and tuples with start and end times of the phase (values)

    include_start: bool, optional
        ``True`` to include data from the beginning of the recording to the start of the first phase as the
        first phase (this phase will be named "Start"), ``False`` to discard this data. Default: ``False``


    Returns
    -------
    dict
        dictionary with data split into different phases


    Examples
    --------
    >>> from biopsykit.utils.data_processing import split_data
    >>> # read pandas dataframe from csv file and split data based on time interval dictionary
    >>> data = pd.read_csv("path-to-file.csv")
    >>> # Example 1: define time intervals (start and end) of the different recording phases as dictionary
    >>> time_intervals = {"Part1": ("09:00", "09:30"), "Part2": ("09:30", "09:45"), "Part3": ("09:45", "10:00")}
    >>> data_dict = split_data(data=data, time_intervals=time_intervals)
    >>> # Example 2: define time intervals as pandas Series. Here, only start times of the are required, it is assumed
    >>> # that the phases are back to back
    >>> time_intervals = pd.Series(data=["09:00", "09:30", "09:45", "10:00"], index=["Part1", "Part2", "Part3", "End"])
    >>> data_dict = split_data(data=data, time_intervals=time_intervals)
    >>>
    >>> # Example: Get Part 2 of data_dict
    >>> print(data_dict['Part2'])

    """
    _assert_is_dtype(time_intervals, (pd.DataFrame, pd.Series, dict))

    if isinstance(time_intervals, pd.DataFrame):
        if len(time_intervals) > 1:
            raise ValueError("Only dataframes with 1 row allowed!")
        time_intervals = time_intervals.iloc[0]

    if isinstance(time_intervals, pd.Series):
        time_intervals = _split_data_series(data, time_intervals, include_start)
    elif include_start:
        time_intervals["Start"] = (
            data.index[0].to_pydatetime().time(),
            list(time_intervals.values())[0][0],
        )

    data_dict = {name: data.between_time(*start_end) for name, start_end in time_intervals.items()}
    data_dict = {name: data for name, data in data_dict.items() if not data.empty}
    return data_dict


def exclude_subjects(
    excluded_subjects: Union[Sequence[str], Sequence[int]], index_name: Optional[str] = "subject", **kwargs
) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """Exclude subjects from dataframes.

    This function can be used to exclude subject IDs for later analysis from different kinds of dataframes, such as:

    * dataframes with subject condition information
      (:obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame`)
    * dataframes with time log information
    * dataframes with (processed) data (e.g., :obj:`biopsykit.utils.datatype_helper.SalivaRawDataFrame` or
      :obj:`~biopsykit.utils.datatype_helper.MeanSeDataFrame`)

    All dataframes can be supplied at once via ``**kwargs``.

    Parameters
    ----------
    excluded_subjects : list of str or int
        list with subjects IDs to be excluded
    index_name : str, optional
        name of dataframe index level with subject IDs. Default: "subject"
    **kwargs :
        data to be cleaned as key-value pairs

    Returns
    -------
    :class:`~pandas.DataFrame` or dict of such
        dictionary with cleaned versions of the dataframes passed to the function via ``**kwargs``
        or dataframe if function was only called with one single dataframe

    """
    cleaned_data: Dict[str, pd.DataFrame] = {}

    for key, data in kwargs.items():
        _assert_is_dtype(data, pd.DataFrame)
        if index_name in data.index.names:
            level_values = data.index.get_level_values(index_name)
            if (level_values.dtype == object and all(isinstance(s, str) for s in excluded_subjects)) or (
                level_values.dtype == int and all(isinstance(s, int) for s in excluded_subjects)
            ):
                cleaned_data[key] = _exclude_single_subject(data, excluded_subjects, index_name, key)
            else:
                raise ValueError(f"{key}: dtypes of index and subject ids to be excluded do not match!")
        else:
            raise ValueError(f"No '{index_name}' level in index!")
    if len(cleaned_data) == 1:
        cleaned_data = list(cleaned_data.values())[0]
    return cleaned_data


def _exclude_single_subject(
    data: pd.DataFrame, excluded_subjects: Union[Sequence[str], Sequence[int]], index_name: str, dataset_name: str
) -> pd.DataFrame:
    # dataframe index and subjects are both strings or both integers
    try:
        if isinstance(data.index, pd.MultiIndex):
            # MultiIndex => specify index level
            return data.drop(index=excluded_subjects, level=index_name)
        # Regular Index
        return data.drop(index=excluded_subjects)
    except KeyError:
        warnings.warn(f"Not all subjects of {excluded_subjects} exist in '{dataset_name}'!")
        return data


def normalize_to_phase(subject_data_dict: SubjectDataDict, phase: Union[str, pd.DataFrame]) -> SubjectDataDict:
    """Normalize time series data per subject to the phase specified by ``normalize_to``.

    The result is the relative change (of, for example, heart rate) compared to the mean value in ``phase``.

    Parameters
    ----------
    subject_data_dict : :class:`~biopsykit.utils.datatype_helper.SubjectDataDict`
        ``SubjectDataDict``, i.e., a dictionary with a :class:`~biopsykit.utils.datatype_helper.PhaseDict`
        for each subject
    phase : str or :class:`~pandas.DataFrame`
        phase to normalize all other data to. If ``phase`` is a string then it is interpreted as the name of a phase
        present in ``subject_data_dict``. If ``phase`` is a DataFrame then the data will be normalized (per subject)
        to the mean value of the DataFrame.

    Returns
    -------
    dict
        dictionary with normalized data per subject

    """
    _assert_is_dtype(phase, (str, pd.DataFrame))
    dict_subjects_norm = {}
    for subject_id, data in subject_data_dict.items():
        bl_mean = data[phase].mean() if isinstance(phase, str) else phase.mean()
        dict_subjects_norm[subject_id] = {p: (df - bl_mean) / bl_mean * 100 for p, df in data.items()}

    return dict_subjects_norm


def resample_sec(data: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    """Resample input data to a frequency of 1 Hz.

    .. note::
        For resampling the index of ``data`` either be has to be a :class:`~pandas.DatetimeIndex`
        or a :class:`~pandas.Index` with time information in seconds.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        data to resample. Index of data needs to be a :class:`~pandas.DatetimeIndex`


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with data resampled to 1 Hz


    Raises
    ------
    ValueError
        If ``data`` is not a DataFrame or Series

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))

    column_name = data.columns if isinstance(data, pd.DataFrame) else [data.name]

    if isinstance(data.index, pd.DatetimeIndex):
        x_old = np.array((data.index - data.index[0]).total_seconds())
    else:
        x_old = np.array(data.index - data.index[0])
    x_new = np.arange(1, np.ceil(x_old[-1]) + 1)
    data = sanitize_input_1d(data)

    interpol_f = interpolate.interp1d(x=x_old, y=data, fill_value="extrapolate")
    return pd.DataFrame(interpol_f(x_new), index=pd.Index(x_new, name="time"), columns=column_name)


def resample_dict_sec(
    data_dict: Dict[str, Any],
) -> Dict[str, Any]:
    """Resample all data in the dictionary to 1 Hz data.

    This function recursively looks for all dataframes in the dictionary and resamples data to 1 Hz using
    :func:`~biopsykit.utils.data_processing.resample_sec`.


    Parameters
    ----------
    data_dict : dict
        nested dictionary with data to be resampled


    Returns
    -------
    dict
        nested dictionary with data resampled to 1 Hz


    See Also
    --------
    :func:`~biopsykit.utils.data_processing.resample_sec`
        resample dataframe to 1 Hz

    """
    result_dict = {}
    for key, value in data_dict.items():
        if isinstance(value, (pd.DataFrame, pd.Series)):
            result_dict[key] = resample_sec(value)
        elif isinstance(value, dict):
            result_dict[key] = resample_dict_sec(value)
        else:
            raise TypeError("Invalid input format!")
    return result_dict


def select_dict_phases(subject_data_dict: SubjectDataDict, phases: Union[str, Sequence[str]]) -> SubjectDataDict:
    """Select specific phases from :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`.

    Parameters
    ----------
    subject_data_dict : :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`
        ``SubjectDataDict``, i.e. a dictionary with :obj:`~biopsykit.utils.datatype_helper.PhaseDict` for each subject
    phases : list of str
        list of phases to select

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`
        ``SubjectDataDict`` containing only the phases of interest

    """
    is_subject_data_dict(subject_data_dict)
    if isinstance(phases, str):
        phases = [phases]
    return {
        subject: {phase: dict_subject[phase] for phase in phases} for subject, dict_subject in subject_data_dict.items()
    }


def rearrange_subject_data_dict(
    subject_data_dict: SubjectDataDict,
) -> StudyDataDict:
    """Rearrange ``SubjectDataDict`` to ``StudyDataDict``.

    A :obj:`~biopsykit.utils.datatype_helper.StudyDataDict` is constructed from a
    :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict` by swapping outer (subject IDs) and inner
    (phase names) dictionary keys.

    The **input** needs to be a :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`,
    a nested dictionary in the following format:

    | {
    |   "subject1" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
    |   "subject2" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
    |   ...
    | }

    The **output** format will be the following:

    | {
    |   "phase_1" : { "subject1" : dataframe, "subject2" : dataframe, ... },
    |   "phase_2" : { "subject1" : dataframe, "subject2" : dataframe, ... },
    |   ...
    | }


    Parameters
    ----------
    subject_data_dict : :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`
        ``SubjectDataDict``, i.e. a dictionary with data from multiple subjects, each containing data from
        multiple phases (in form of a :obj:`~biopsykit.utils.datatype_helper.PhaseDict`)


    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`
        rearranged ``SubjectDataDict``

    """
    dict_flipped = {}
    phases = [np.array(dict_phase.keys()) for dict_phase in subject_data_dict.values()]
    if not all(phases[0] == p for p in phases):
        raise ValueError(
            "Error rearranging the dictionary! Not all 'PhaseDict's have the same phases. "
            "To rearrange the 'SubjectDataDict', "
            "the dictionaries of all subjects need to have the exact same phases!"
        )

    for subject, phase_dict in subject_data_dict.items():
        for phase, df in phase_dict.items():
            dict_flipped.setdefault(phase, dict.fromkeys(subject_data_dict.keys()))
            dict_flipped[phase][subject] = df

    return dict_flipped


def cut_phases_to_shortest(study_data_dict: StudyDataDict, phases: Optional[Sequence[str]] = None) -> StudyDataDict:
    """Cut time-series data to shortest duration of a subject in each phase.

    To overlay time-series data from multiple subjects in an `ensemble plot` it is beneficial if all data have
    the same length. For that reason, data can be cut to the same length using this function.

    Parameters
    ----------
    study_data_dict : :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`
        ``StudyDataDict``, i.e. a dictionary with data from multiple phases, each phase containing data from
        different subjects.
    phases : list of str, optional
        list of phases if only a subset of phases should be cut or ``None`` to cut all phases.
        Default: ``None``

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`
        ``StudyDataDict`` with data cut to the shortest duration in each phase

    """
    is_study_data_dict(study_data_dict)

    if phases is None:
        phases = study_data_dict.keys()

    dict_cut = {}
    for phase in phases:
        min_dur = min(len(df) for df in study_data_dict[phase].values())
        dict_cut[phase] = {subject: df.iloc[0:min_dur].copy() for subject, df in study_data_dict[phase].items()}

    is_study_data_dict(study_data_dict)
    return dict_cut


def merge_study_data_dict(
    study_data_dict: StudyDataDict, dict_levels: Optional[Sequence[str]] = None
) -> MergedStudyDataDict:
    """Merge inner dictionary level of :obj:`~biopsykit.utils.datatype_helper.StudyDataDict` into one dataframe.

    This function removes the inner level of the nested ``StudyDataDict`` by merging data from all subjects
    into one dataframe for each phase.

    .. note::
        To merge data from different subjects into one dataframe the data are all expected to have the same length!
        If this is not the case, all data needs to be cut to equal length first, e.g. using
        :func:`~biopsykit.utils.data_processing.cut_phases_to_shortest`.


    Parameters
    ----------
    study_data_dict : :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`
        ``StudyDataDict``, i.e. a dictionary with data from multiple phases, each phase containing data from
        different subjects.
    dict_levels : list of str
        list with names of dictionary levels.


    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`
        ``MergedStudyDataDict`` with data of all subjects merged into one dataframe for each phase

    """
    is_study_data_dict(study_data_dict)

    if dict_levels is None:
        dict_levels = ["phase", "subject"]

    dict_merged = {}
    for phase, dict_phase in study_data_dict.items():
        _assert_dataframes_same_length(list(dict_phase.values()))
        df_merged = pd.concat(dict_phase, names=dict_levels[1:], axis=1)
        df_merged.columns = df_merged.columns.droplevel(-1)
        dict_merged[phase] = df_merged

    is_merged_study_data_dict(dict_merged)
    return dict_merged


def split_dict_into_subphases(
    data_dict: Dict[str, Any],
    subphases: Dict[str, int],
) -> Union[Dict[str, Dict[str, Any]]]:
    """Split dataframes in a nested dictionary into subphases.

    By further splitting a dataframe into subphases a new dictionary level is created. The new dictionary level
    then contains the subphases with their data.

    .. note::
        If the duration of the last subphase is unknown (e.g., because it has variable length) this can be
        indicated by setting the duration of this subphase to 0.
        The duration of this subphase will then be inferred from the data.

    Parameters
    ----------
    data_dict : dict
        dictionary with an arbitrary number of outer level (e.g., conditions, phases, etc.) as keys and
        dataframes with data to be split into subphases as values
    subphases : dict
        dictionary with subphase names (keys) and subphase durations (values) in seconds

    Returns
    -------
    dict
        dictionary where each dataframe in the dictionary is split into the subphases specified by ``subphases``

    """
    result_dict = {}
    for key, value in data_dict.items():
        _assert_is_dtype(value, (dict, pd.DataFrame))
        if isinstance(value, dict):
            # nested dictionary
            result_dict[key] = split_dict_into_subphases(value, subphases)
        else:
            subphase_times = get_subphase_durations(value, subphases)
            subphase_dict = {}
            for subphase, times in zip(subphases.keys(), subphase_times):
                if isinstance(value.index, pd.DatetimeIndex):
                    # slice the current subphase by dropping the preceding subphases
                    value_cpy = value.drop(value.first(f"{times[0]}s").index)
                    value_cpy = value_cpy.first(f"{times[1] - times[0]}s")
                    subphase_dict[subphase] = value_cpy
                else:
                    subphase_dict[subphase] = value.iloc[times[0] : times[1]]
            result_dict[key] = subphase_dict
    return result_dict


def get_subphase_durations(
    data: pd.DataFrame, subphases: Dict[str, Union[int, Tuple[int, int]]]
) -> Sequence[Tuple[int, int]]:
    """Compute subphase durations from dataframe.

    The subphases can be specified in two different ways:

    * If the dictionary entries in ``subphases`` are integer, it's assumed that subphases are consecutive,
      i.e., each subphase begins right after the previous one, and the entries indicate the *durations* of each
      subphase. The start and end times of each subphase will then be computed from the subphase durations.
    * If the dictionary entries in ``subphases`` are tuples, it's assumed that the start and end times of each
      subphase are directly provided.

    .. note::
        If the duration of the last subphase is unknown (e.g., because it has variable length) this can be
        indicated by setting the duration of this subphase to 0.
        The duration of this subphase will then be inferred from the data.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with data from one phase. Used to compute the duration of the last subphase if this subphase
        is expected to have variable duration.
    subphases : dict
        dictionary with subphase names as keys and subphase durations (as integer) or start and end
        times (as tuples of integer) as values in seconds

    Returns
    -------
    list
        list with start and end times of each subphase in seconds relative to beginning of the phase


    Examples
    --------
    >>> from biopsykit.utils.data_processing import get_subphase_durations
    >>> # Option 1: Subphases consecutive, subphase durations provided
    >>> get_subphase_durations(data, {"Start": 60, "Middle": 120, "End": 60})
    >>> # Option 2: Subphase start and end times provided
    >>> get_subphase_durations(data, {"Start": (0, 50), "Middle": (60, 160), "End": (180, 240)})

    """
    subphase_durations = np.array(list(subphases.values()))
    if subphase_durations.ndim == 1:
        # 1d array => subphase values are integer => they are consecutive and each entry is the duration
        # of the subphase, so the start and end times of each subphase must be computed
        times_cum = np.cumsum(subphase_durations)
        if subphase_durations[-1] == 0:
            # last subphase has duration 0 => end of last subphase is length of dataframe
            times_cum[-1] = len(data)
        subphase_times = list(zip([0, *list(times_cum)], times_cum))
    else:
        # 2d array => subphase values are tuples => start end end time of each subphase are already provided and do
        # not need to be computed
        subphase_times = subphase_durations
    return subphase_times


def add_subject_conditions(
    data: pd.DataFrame, condition_list: Union[SubjectConditionDict, SubjectConditionDataFrame]
) -> pd.DataFrame:
    """Add subject conditions to dataframe.

    This function expects a dataframe with data from multiple subjects and information on which subject
    belongs to which condition.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe where new index level ``condition`` with subject conditions should be added to
    condition_list : ``SubjectConditionDict`` or ``SubjectConditionDataFrame``
        :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` or
        :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` with information on which subject belongs to
        which condition


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with new index level ``condition`` indicating which subject belongs to which condition

    """
    if is_subject_condition_dataframe(condition_list, raise_exception=False):
        condition_list = condition_list.groupby("condition").groups
    is_subject_condition_dict(condition_list)
    return pd.concat({cond: data.loc[subjects] for cond, subjects in condition_list.items()}, names=["condition"])


def split_subject_conditions(
    data_dict: Dict[str, Any], condition_list: Union[SubjectConditionDict, SubjectConditionDataFrame]
) -> Dict[str, Dict[str, Any]]:
    """Split dictionary with data based on conditions subjects were assigned to.

    This function adds a new outer dictionary level with the different conditions as keys and dictionaries
    belonging to the conditions as values. For that, it expects a dictionary with data from multiple subjects and
    information on which subject belongs to which condition.


    Parameters
    ----------
    data_dict : dict
        (nested) dictionary with data which should be split based on the conditions subjects belong to
    condition_list : ``SubjectConditionDict`` or ``SubjectConditionDataFrame``
        :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict` or
        :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame` with information on which subject belongs to
        which condition


    Returns
    -------
    dict
        dictionary with additional outer level indicating which subject belongs to which condition

    """
    if is_subject_condition_dataframe(condition_list, raise_exception=False):
        condition_list = condition_list.groupby("condition").groups
    is_subject_condition_dict(condition_list)
    return {cond: _splits_subject_conditions(data_dict, subjects) for cond, subjects in condition_list.items()}


def _splits_subject_conditions(data_dict: Dict[str, Any], subject_list: Sequence[str]):
    _assert_is_dtype(data_dict, (dict, pd.DataFrame))
    if isinstance(data_dict, pd.DataFrame):
        return data_dict[subject_list]
    return {key: _splits_subject_conditions(value, subject_list) for key, value in data_dict.items()}


def mean_per_subject_dict(data: Dict[str, Any], dict_levels: Sequence[str], param_name: str) -> pd.DataFrame:
    """Compute mean values of time-series data from a nested dictionary.

    This function computes the mean value of time-series data in a nested dictionary per subject and combines it into
    a joint dataframe. The dictionary will be traversed recursively and can thus have arbitrary depth.
    The most inner level must contain dataframes with time-series data of which mean values will be computed.
    The names of the dictionary levels are specified by ``dict_levels``.


    Parameters
    ----------
    data: dict
        nested dictionary with data on which mean should be computed. The number of nested levels must match the
        number of levels specified in ``dict_levels``.
    dict_levels : list of str
        list with names of dictionary levels.
    param_name : str
        type of data of which mean values will be computed from.
        This will also be the column name in the resulting dataframe.


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with ``dict_levels`` as index levels and mean values of time-series data as column values

    """
    result_data = {}

    one_col_df = False
    for key, value in data.items():
        _assert_is_dtype(value, (dict, pd.DataFrame))
        if isinstance(value, dict):
            if len(dict_levels) <= 1:
                raise ValueError("Invalid number of 'dict_levels' specified!")
            # nested dictionary
            key_len = 1 if isinstance(key, (str, int)) else len(key)
            result_data[key] = mean_per_subject_dict(value, dict_levels[key_len:], param_name)
        else:
            value.columns.name = "subject"
            if len(value.columns) == 1:
                one_col_df = True
            df = pd.DataFrame(value.mean(axis=0), columns=[param_name])
            result_data[key] = df

    key_lengths = list({1 if isinstance(k, (str, int)) else len(k) for k in result_data})
    if len(key_lengths) != 1:
        raise ValueError("Inconsistent dictionary key lengths!")
    key_lengths = key_lengths[0]
    names = dict_levels[0:key_lengths]
    if isinstance(names, str):
        names = [names]
    ret = pd.concat(result_data, names=names)
    if one_col_df:
        ret.index = ret.index.droplevel(-1)
    return ret


def mean_se_per_phase(data: pd.DataFrame) -> MeanSeDataFrame:
    """Compute mean and standard error over all subjects in a dataframe.

    .. note::
        The dataframe in ``data`` is expected to have a :class:`~pandas.MultiIndex` with at least two levels,
        one of the levels being the level "subject"!

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with :class:`~pandas.MultiIndex` from which to compute mean and standard error

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with mean and standard error over all subjects

    """
    # expect dataframe to have at least 2 levels, one of it being "subject"
    _assert_has_multiindex(data, expected=True, nlevels=2, nlevels_atleast=True)
    _assert_has_index_levels(data, ["subject"], match_atleast=True)

    # group by all columns except the "subject" column
    group_cols = list(data.index.names)
    group_cols.remove("subject")

    data = data.groupby(group_cols, sort=False).agg([np.mean, se])
    is_mean_se_dataframe(data)

    return _MeanSeDataFrame(data)



"""Functions for performing regression analysis."""
from typing import Union

import pandas as pd
import pingouin as pg

__all__ = ["stepwise_backwards_linear_regression"]


def stepwise_backwards_linear_regression(predictors: pd.DataFrame, dv: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    """Perform a stepwise backwards linear regression.

    The stepwise backwards linear regression is performed iteratively by running a linear regression on the predictors
    and removing the predictor with the highest p-value. The best-fitting model is obtained by choosing the regression
    model with the highest adjusted R-squared value.

    Parameters
    ----------
    predictors : :class:`pandas.DataFrame`
        Dataframe with predictors for the regression.
    dv : :class:`pandas.DataFrame` or :class:`pandas.Series`
        Dependent variable for the regression.

    Returns
    -------
    :class:`pandas.DataFrame`
        Best-fitted regression model.

    """
    # dv must only be a single column
    if dv.ndim > 1:
        if dv.shape[-1] != 1:
            raise ValueError("dv must be a single column")
        dv = dv.iloc[:, 0]

    list_adj_r2 = []
    list_reg_models = []
    while len(predictors.columns) > 0:
        reg_results = _lin_reg(predictors, dv)
        # drop the most correlated predictor
        drop_col = reg_results["pval"].idxmax()
        predictors = predictors.drop(columns=drop_col)
        adj_r2 = reg_results["adj_r2"].iloc[0]
        list_adj_r2.append(adj_r2)
        list_reg_models.append(reg_results)

    # get the best regression model, i.e. the model with the highest adjusted R2
    return list_reg_models[list_adj_r2.index(max(list_adj_r2))]


def _lin_reg(predictors: pd.DataFrame, dv: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    return pg.linear_regression(predictors, dv).set_index("names").drop(index="Intercept")
